/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author     Raphael Grimm
 * @date       2015
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "DepthImageProviderDynamicSimulation.h"

#include <boost/algorithm/clamp.hpp>

#include <Eigen/Geometry>

#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/SoInteraction.h>

#include <pcl/common/transforms.h>
#include <pcl/compression/organized_pointcloud_conversion.h>


using namespace visionx;
using namespace VirtualRobot;

namespace armarx
{
    void DepthImageProviderDynamicSimulation::onInitComponent()
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onInitComponent";
        baseline = getProperty<float>("BaseLine").getValue();
        robotName = getProperty<std::string>("RobotName").getValue();
        camNodeName = getProperty<std::string>("RobotNodeCamera").getValue();

        zNear = getProperty<float>("DistanceZNear").getValue();
        zFar = getProperty<float>("DistanceZFar").getValue();
        vertFov = getProperty<float>("FOV").getValue();
        nanValue = getProperty<float>("NanValue").getValue();
        // focalLength is set in onInitImageProvider


        drawPointCloud = getProperty<bool>("DrawPointCloud").getValue();
        if (drawPointCloud)
        {
            lastUpdate = std::chrono::high_resolution_clock::now() - std::chrono::seconds {pointCloudDrawDelay};
            debugDrawerTopicName = getProperty<std::string>("DrawPointCloud_DebugDrawerTopic").getValue();
            offeringTopic(debugDrawerTopicName);
            pointCloudDrawDelay = getProperty<unsigned int>("DrawPointCloud_DrawDelay").getValue();
            pointCloudPointSize = getProperty<float>("DrawPointCloud_PointSize").getValue();

            pointSkip =  getProperty<std::size_t>("DrawPointCloud_PointSkip").getValue();

            clipDrawnCloudXLo = getProperty<float>("DrawPointCloud_ClipXLo").getValue();
            clipDrawnCloudYLo = getProperty<float>("DrawPointCloud_ClipYLo").getValue();
            clipDrawnCloudZLo = getProperty<float>("DrawPointCloud_ClipZLo").getValue();

            clipDrawnCloudXHi = getProperty<float>("DrawPointCloud_ClipXHi").getValue();
            clipDrawnCloudYHi = getProperty<float>("DrawPointCloud_ClipYHi").getValue();
            clipDrawnCloudZHi = getProperty<float>("DrawPointCloud_ClipZHi").getValue();

            clipPointCloud = getProperty<bool>("DrawPointCloud_ClipPoints").getValue();
        }
        noise = getProperty<float>("Noise").getValue();
        distribution = std::normal_distribution<double>(0, noise);
        if (noise > 0)
        {
            randValues.resize(1000, 0);
            for (float & val : randValues)
            {
                val = distribution(generator);
            }
        }



        ImageProvider::onInitComponent();
        CapturingPointCloudProvider::onInitComponent();
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onInitComponent";
    }

    void DepthImageProviderDynamicSimulation::onConnectComponent()
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onConnectComponent";

        if (drawPointCloud)
        {
            debugDrawer = getTopic<DebugDrawerInterfacePrx>(debugDrawerTopicName);
        }

        ImageProvider::onConnectComponent();
        CapturingPointCloudProvider::onConnectComponent();
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onConnectComponent";
    }

    void DepthImageProviderDynamicSimulation::onExitComponent()
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onExitComponent";
        CapturingPointCloudProvider::onExitComponent();
        ImageProvider::onExitComponent();
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onExitComponent";
    }

    void DepthImageProviderDynamicSimulation::onStartCapture(float frameRate)
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onStartCapture";
        // ensure that all data is loaded
        if (simVisu)
        {
            simVisu->synchronizeVisualizationData();
        }
        while (!simVisu->getRobot(robotName))
        {
            usleep(100000);
            simVisu->synchronizeVisualizationData();
            ARMARX_INFO << deactivateSpam(3) << "Waiting for visu";
        }

        boost::mutex::scoped_lock lock(captureMutex);
        ScopedRecursiveLockPtr l = simVisu->getScopedLock();

        // some checks...
        if (!simVisu || !simVisu->getVisualization())
        {
            ARMARX_ERROR << "No physics visualization scene";
        }
        offscreenRenderer.reset(CoinVisualizationFactory::createOffscreenRenderer(width, height));
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onStartCapture";
    }

    CameraParameters DepthImageProviderDynamicSimulation::getDepthCameraIntrinsics(const Ice::Current&)
    {
        visionx::CameraParameters params;
        params.distortion = {0, 0, 0};
        params.focalLength = {focalLength, focalLength};
        params.height = height;
        params.principalPoint = {width / 2.f, height / 2.f};
        params.rotation = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}};
        params.translation = {0.075, 0, 0};
        params.width = width;

        return params;
    }

    void DepthImageProviderDynamicSimulation::onExitCapturingPointCloudProvider()
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onExitCapturingPointCloudProvider";

        rgbImage.reset();
        dImage.reset();
        dFloatImage.reset();
        imgPtr[0] = nullptr;
        imgPtr[1] = nullptr;
        imgPtrFlt[0] = nullptr;

        if (simVisu)
        {
            getArmarXManager()->removeObjectBlocking(simVisu);
        }
        simVisu = nullptr;
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onExitCapturingPointCloudProvider";
    }

    void DepthImageProviderDynamicSimulation::onInitImageProvider()
    {
        ARMARX_VERBOSE_S << "DepthImageProviderDynamicSimulation::onInitImageProvider";

        const auto imgSz = getProperty<ImageDimension>("ImageSize").getValue();
        height = imgSz.height;
        width = imgSz.width;

        if (!getProperty<bool>("FloatImageMode").getValue())
        {
            setImageFormat(visionx::ImageDimension(imgSz), visionx::eRgb);
            setNumberImages(2);

            rgbImage.reset(new CByteImage {static_cast<int>(width), static_cast<int>(height), CByteImage::eRGB24});
            dImage.reset(new CByteImage {static_cast<int>(width), static_cast<int>(height), CByteImage::eRGB24});
            imgPtr[0] = rgbImage.get();
            imgPtr[1] = dImage.get();
        }
        else
        {
            setImageFormat(visionx::ImageDimension(imgSz), visionx::eFloat1Channel);
            setNumberImages(1);

            dFloatImage.reset(new CFloatImage {static_cast<int>(width), static_cast<int>(height), 1});
            imgPtrFlt[0] = dFloatImage.get();
        }

        // verticalFOV = 2 * atan(height/(2*focalLength))
        focalLength = static_cast<float>(height) / (2 * std::tan(vertFov / 2));

        pointcloud.reset(new pcl::PointCloud<PointT> {static_cast<unsigned int>(width), static_cast<unsigned int>(height)});

        rgbBuffer.resize(width * height);
        depthBuffer.resize(width * height);
        pointCloudBuffer.resize(width * height);

        // init SoDB / Coin3D
        SoDB::init();

        VirtualRobot::init(getName());

        // needed for SoSelection
        SoInteraction::init();

        std::stringstream svName;
        svName << getName() << "_PhysicsWorldVisualization";
        simVisu = Component::create<ArmarXPhysicsWorldVisualization>(getIceProperties(), getName() + "_PhysicsWorldVisualization");
        getArmarXManager()->addObject(simVisu);
        ARMARX_VERBOSE_S << "done DepthImageProviderDynamicSimulation::onInitImageProvider";

        renderTimesCount = 0;
        sumRenderTimes = 0;
        sumRenderTimesSquared = 0;
    }

    bool DepthImageProviderDynamicSimulation::doCapture()
    {
        bool succeeded = true;
        Ice::Long timestamp;
        try
        {
            timestamp = simVisu->getSyncTimestamp();
            simVisu->synchronizeVisualizationData();
            succeeded = render();
        }
        catch (...)
        {
            handleExceptions();
            succeeded = false;
        }

        if (succeeded)
        {
            Eigen::Matrix4f rotate = Eigen::Matrix4f::Identity();
            // rotate to conform to ArmarX standards
            rotate.block(0, 0, 3, 3) = Eigen::AngleAxis<float>(M_PI_2, Eigen::Vector3f::UnitZ()). toRotationMatrix();
            pcl::PointCloud<PointT>::Ptr cloudTransformedPtr(new pcl::PointCloud<PointT>());
            pcl::transformPointCloud(*pointcloud, *cloudTransformedPtr, rotate);
            cloudTransformedPtr->header.stamp = timestamp;
            providePointCloud<PointT>(cloudTransformedPtr);

            if (!getProperty<bool>("FloatImageMode").getValue())
            {
                provideImages(imgPtr);
            }
            else
            {
                provideImages(imgPtrFlt);
            }

            if (drawPointCloud && (std::chrono::high_resolution_clock::now() - lastUpdate > std::chrono::milliseconds {pointCloudDrawDelay}))
            {
                lastUpdate = std::chrono::high_resolution_clock::now();
                DebugDrawer24BitColoredPointCloud dbgPcl;
                dbgPcl.points.reserve(cloudTransformedPtr->size());

                dbgPcl.pointSize = pointCloudPointSize;

                const auto pointCloudBaseNode = simVisu->getRobot(robotName)->getRobotNode(camNodeName);

                if (!pointCloudBaseNode)
                {
                    ARMARX_ERROR_S << deactivateSpam() << robotName << " has no node " << camNodeName
                                   << "\nthe point cloud will be attached to (0,0,0)";
                }

                Eigen::Matrix4f rotate = Eigen::Matrix4f::Identity();
                //                rotate.block(0, 0, 3, 3) = Eigen::AngleAxis<float>(M_PI_2, Eigen::Vector3f::UnitZ()). toRotationMatrix();
                Eigen::Matrix4f transformCamToWorld = (pointCloudBaseNode ? pointCloudBaseNode->getGlobalPose() : Eigen::Matrix4f::Identity()) * rotate;


                assert(static_cast<std::size_t>(width)  == pointcloud->width);
                assert(static_cast<std::size_t>(height) == pointcloud->height);
                for (std::size_t x = 0; x < static_cast<std::size_t>(width); x += pointSkip)
                {
                    for (std::size_t y = 0; y < static_cast<std::size_t>(height); y += pointSkip)
                    {
                        const auto& pclPoint = cloudTransformedPtr->at(x, y);

                        //transform to point in world coord
                        const Eigen::Vector4f pointFromPcl
                        {
                            pclPoint.x, pclPoint.y, pclPoint.z, 1
                        };

                        const Eigen::Vector4f pointInWorld = transformCamToWorld * pointFromPcl ;
                        const Eigen::Vector4f pointInWorldNormalizedW = pointInWorld / pointInWorld(3);

                        DebugDrawer24BitColoredPointCloudElement dbgPoint;

                        dbgPoint.x = pointInWorldNormalizedW(0);
                        dbgPoint.y = pointInWorldNormalizedW(1);
                        dbgPoint.z = pointInWorldNormalizedW(2);

                        //clip to space
                        if (
                            clipPointCloud &&
                            (
                                dbgPoint.x < clipDrawnCloudXLo || dbgPoint.x > clipDrawnCloudXHi ||
                                dbgPoint.y < clipDrawnCloudYLo || dbgPoint.y > clipDrawnCloudYHi ||
                                dbgPoint.z < clipDrawnCloudZLo || dbgPoint.z > clipDrawnCloudZHi
                            )
                        )
                        {
                            continue;
                        }

                        dbgPoint.color.r = pclPoint.r;
                        dbgPoint.color.g = pclPoint.g;
                        dbgPoint.color.b = pclPoint.b;

                        dbgPcl.points.emplace_back(dbgPoint);
                    }
                }

                debugDrawer->set24BitColoredPointCloudVisu("PointClouds", getName() + "::PointCloud", dbgPcl);
                //                debugDrawer->setPoseVisu("PointClouds", getName() + "::PointCloudBaseNode", new Pose { (pointCloudBaseNode ? pointCloudBaseNode->getGlobalPose() : Eigen::Matrix4f::Identity()) });
            }
        }
        return succeeded;
    }

    bool DepthImageProviderDynamicSimulation::render()
    {
        //render images

        if (!simVisu || !simVisu->getRobot(robotName))
        {
            ARMARX_WARNING << deactivateSpam() << " no visu or no robot";
            usleep(100000);
            return false;
        }

        const auto cameraNode = simVisu->getRobot(robotName)->getRobotNode(camNodeName);

        if (!cameraNode)
        {
            ARMARX_ERROR << deactivateSpam() << "No camera node:" << camNodeName;
            return false;
        }
        bool renderOK = false;
        {
            ScopedRecursiveLockPtr l = simVisu->getScopedLock();
            auto start = IceUtil::Time::now();

            renderOK = CoinVisualizationFactory::renderOffscreenRgbDepthPointcloud(offscreenRenderer.get(), cameraNode, simVisu->getVisualization(), width, height,
                       true, rgbBuffer,
                       true, depthBuffer,
                       true,
                       pointCloudBuffer, zNear, zFar, vertFov, nanValue);

            ARMARX_DEBUG << deactivateSpam(1) << "Offscreen rendering took: " << (IceUtil::Time::now() - start).toMilliSecondsDouble() << " ms";
        }
        if (renderOK)
        {
            //get images + cloud
            boost::mutex::scoped_lock lock(captureMutex);

            auto start = std::chrono::high_resolution_clock::now();

            auto randValuesSize = randValues.size();
            int noiseOffset = noise > 0 ? rand() % randValuesSize : 0;
            bool floatImageMode = getProperty<bool>("FloatImageMode").getValue();

            float* depthBufferEndOfRow = depthBuffer.data() + width - 1;
            if (floatImageMode)
            {
                float* dFloatImageRow = dFloatImage->pixels;
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        float value = boost::algorithm::clamp(0.f, depthBufferEndOfRow[-x], static_cast<float>(0xffffff));
                        dFloatImageRow[x] = value;
                    }
                    dFloatImageRow += width;
                    depthBufferEndOfRow += width;
                }
            }
            else
            {
                std::uint8_t* rgbImageRow = rgbImage->pixels;
                std::uint8_t* rgbBufferEndOfRow = rgbBuffer.data() + (width - 1) * 3;
                std::uint8_t* dImageRow = dImage->pixels;
                for (int y = 0; y < height; ++y)
                {
                    for (int x = 0; x < width; ++x)
                    {
                        rgbImageRow[x * 3]     = rgbBufferEndOfRow[-x * 3];
                        rgbImageRow[x * 3 + 1] = rgbBufferEndOfRow[-x * 3 + 1];
                        rgbImageRow[x * 3 + 2] = rgbBufferEndOfRow[-x * 3 + 2];
                        std::int32_t value = boost::algorithm::clamp(0.f, depthBufferEndOfRow[-x], static_cast<float>(0xffffff));
                        dImageRow[x * 3]     = value & 0xff;
                        dImageRow[x * 3 + 1] = (value >> 8) & 0xff;
                        dImageRow[x * 3 + 2] = (value >> 16) & 0xff;
                    }
                    rgbImageRow += width * 3;
                    rgbBufferEndOfRow += width * 3;
                    dImageRow += width * 3;
                    depthBufferEndOfRow += width;
                }
            }

            float* randValuesData = randValues.data();
            std::uint8_t* rgbBufferEntry = rgbBuffer.data();
            Eigen::Vector3f* inputRow = pointCloudBuffer.data();
            PointT* outputEndOfRow = pointcloud->points.data() + (width - 1);
            for (int y = 0; y < height; ++y)
            {
                for (int x = 0; x < width; ++x)
                {
                    PointT& point = outputEndOfRow[-x];
                    Eigen::Vector3f pointCloudBufferPoint = inputRow[x];

                    int noiseIndex = (y * width + x) + noiseOffset;
                    float noiseX = noise > 0.0f ? randValuesData[(noiseIndex) % randValuesSize] : 0.0f;
                    float noiseY = noise > 0.0f ? randValuesData[(noiseIndex + 1) % randValuesSize] : 0.0f;
                    float noiseZ = noise > 0.0f ? randValuesData[(noiseIndex + 2) % randValuesSize] : 0.0f;

                    point.x = pointCloudBufferPoint[0] + noiseX;
                    point.y = pointCloudBufferPoint[1] + noiseY;
                    point.z = pointCloudBufferPoint[2] + noiseZ;
                    point.r = rgbBufferEntry[0];
                    point.g = rgbBufferEntry[1];
                    point.b = rgbBufferEntry[2];

                    rgbBufferEntry += 3;
                }
                inputRow += width;
                outputEndOfRow += width;
            }

            auto end = std::chrono::high_resolution_clock::now();
            std::int64_t timeDiff = std::chrono::duration_cast<std::chrono::microseconds>(end - start).count();
            sumRenderTimes += timeDiff;
            sumRenderTimesSquared += timeDiff * timeDiff;
            renderTimesCount += 1;

            if (renderTimesCount > 256)
            {
                double averageTime = sumRenderTimes * 1.0e-3 / renderTimesCount;
                double averageSquaredTime = sumRenderTimesSquared * 1.0e-6 / renderTimesCount;
                double stddevTime = std::sqrt(averageSquaredTime - averageTime * averageTime);

                ARMARX_DEBUG << "Copying depth buffer data took: " << averageTime << "ms +- " << stddevTime << "ms";

                renderTimesCount = 0;
                sumRenderTimes = 0;
                sumRenderTimesSquared = 0;
                renderTimesCount = 0;
            }
        }
        return renderOK;
    }
}
