/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::KinematicUnitDynamicSimulation
 * @author     Nikolaus ( vahrenkamp at kit dot edu )
 * @date       2013
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "KinematicUnitDynamicSimulation.h"

#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/RobotNodeSet.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/VirtualRobotException.h>
#include <VirtualRobot/RuntimeEnvironment.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>

#include <algorithm>

using namespace std;
using namespace armarx;


void KinematicUnitDynamicSimulation::onInitKinematicUnit()
{
    ARMARX_VERBOSE << "Init with RNS " << robotNodeSetName << endl;

    for (std::vector<VirtualRobot::RobotNodePtr>::iterator it = robotNodes.begin(); it != robotNodes.end(); it++)
    {
        std::string jointName = (*it)->getName();
        ARMARX_VERBOSE << "* " << jointName << flush;
        nodeNames.insert(jointName);
    }

    if (robot)
    {
        robotTopicName = "Simulator_Robot_";
        robotTopicName += robot->getName();
        robotName = robot->getName();
        ARMARX_INFO << "Simulator Robot topic: " << robotTopicName << endl;
        usingTopic(robotTopicName);
    }
    else
    {
        ARMARX_WARNING << "No robot loaded..." << endl;
    }

    mapVelToPosVel = getProperty<bool>("MapVelocityToPositionVelocity").getValue();

    simulatorPrxName = getProperty<std::string>("SimulatorName").getValue();
    usingProxy(simulatorPrxName);
}

void KinematicUnitDynamicSimulation::onStartKinematicUnit()
{
    // subscribe topic in order to receive the simulator reports
    ARMARX_INFO << "Using simulator proxy: " << simulatorPrxName << flush;
    simulatorPrx = getProxy<SimulatorInterfacePrx>(simulatorPrxName);

    continuousMode = getProperty<bool>("ContinuousMode").getValue();


    if (!continuousMode)
    {
        // periodic task setup
        cycleTime = 1000.f / getProperty<float>("ControlFrequency").getValue();

        if (execTask)
        {
            execTask->stop();
        }

        execTask = new PeriodicTask<KinematicUnitDynamicSimulation>(this, &KinematicUnitDynamicSimulation::periodicExec, cycleTime, false, "KinematicUnitDynamicSimulation");
        execTask->start();
        execTask->setDelayWarningTolerance(100);
    }
}


void KinematicUnitDynamicSimulation::onExitKinematicUnit()
{
    if (execTask)
    {
        execTask->stop();
    }
}

void KinematicUnitDynamicSimulation::reportJointAngles(const NameValueMap& actualJointAngles, Ice::Long timestamp, bool aValueChanged, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    if (!aValueChanged)
    {
        return;
    }

    if (actualJointAngles.size() == 0)
    {
        return;
    }

    // check which joints are covered
    NameValueMap res;

    for (NameValueMap::const_iterator it = actualJointAngles.begin(); it != actualJointAngles.end(); it++)
    {
        if (nodeNames.find(it->first) != nodeNames.end())
        {
            res[it->first] = it->second;
        }

        // store current values
        currentJointValues[it->first] = it->second;
    }

    //ARMARX_DEBUG << "reportJointAngles, first joint: " << actualJointAngles.begin()->first <<":"<< actualJointAngles.begin()->second << endl;

    // send to KinematicUnitPrx
    if (res.size() > 0)
    {
        listenerPrx->reportJointAngles(res, timestamp, true);
    }
}

void KinematicUnitDynamicSimulation::reportJointVelocities(const NameValueMap& actualJointVelocities, bool aValueChanged, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);
    auto time = TimeUtil::GetTime().toMicroSeconds();
    if (!aValueChanged)
    {
        return;
    }

    // check which jonts are covered
    NameValueMap res;

    for (NameValueMap::const_iterator it = actualJointVelocities.begin(); it != actualJointVelocities.end(); it++)
    {
        if (nodeNames.find(it->first) != nodeNames.end())
        {
            res[it->first] = it->second;
        }
    }

    ARMARX_DEBUG << "reportJointVelocities first joint: " << actualJointVelocities.begin()->first << ":" << actualJointVelocities.begin()->second << endl;

    // send to KinematicUnitPrx
    if (res.size() > 0)
    {
        listenerPrx->reportJointVelocities(res, time, true);
    }
}


void KinematicUnitDynamicSimulation::reportJointTorques(const NameValueMap& actualJointTorques, bool aValueChanged, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    if (!aValueChanged)
    {
        return;
    }

    // check which jonts are covered
    NameValueMap res;

    for (NameValueMap::const_iterator it = actualJointTorques.begin(); it != actualJointTorques.end(); it++)
    {
        if (nodeNames.find(it->first) != nodeNames.end())
        {
            res[it->first] = it->second;
        }
    }

    ARMARX_DEBUG << "actualJointTorques first joint: " << actualJointTorques.begin()->first << ":" << actualJointTorques.begin()->second << endl;

    // send to KinematicUnitPrx
    if (res.size() > 0)
    {
        listenerPrx->reportJointTorques(res, IceUtil::Time::now().toMicroSeconds(), true);
    }
}


void KinematicUnitDynamicSimulation::reportPose(const PoseBasePtr& actualPose, bool aValueChanged, const Ice::Current& c)
{
    // nyi
}

void KinematicUnitDynamicSimulation::reportVelocity(const Vector3BasePtr& translationVel, const Vector3BasePtr& rotationVel, bool aValueChanged, const Ice::Current& c)
{
    // nyi
}

NameControlModeMap KinematicUnitDynamicSimulation::getControlModes(const Ice::Current& c)
{
    NameControlModeMap result;

    for (auto it = nodeNames.begin(); it != nodeNames.end(); it++)
    {
        auto c = currentControlModes.find(*it);

        // if not specified, we assume position control
        if (c == currentControlModes.end())
        {
            result[*it] = ePositionControl;
        }
        else
        {
            result[*it] = c->second;
        }
    }

    return result;
}


void KinematicUnitDynamicSimulation::switchControlMode(const NameControlModeMap& targetJointModes, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    //currentControlModes = targetJointModes;
    NameControlModeMap::const_iterator itAng = targetJointModes.begin();

    while (itAng != targetJointModes.end())
    {
        currentControlModes[itAng->first] = itAng->second;
        itAng++;
    }

    listenerPrx->reportControlModeChanged(targetJointModes, IceUtil::Time::now().toMicroSeconds(), true);

}

void KinematicUnitDynamicSimulation::setJointAngles(const NameValueMap& targetJointAngles, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);



    // store current joint angles
    NameValueMap::const_iterator itAng = targetJointAngles.begin();

    while (itAng != targetJointAngles.end())
    {
        currentJointTargets[itAng->first] = itAng->second;
        itAng++;
    }
}

void KinematicUnitDynamicSimulation::setJointVelocities(const NameValueMap& targetJointVelocities, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);
    NameValueMap::const_iterator itVel = targetJointVelocities.begin();

    while (itVel != targetJointVelocities.end())
    {
        currentJointVelTargets[itVel->first] = itVel->second;
        itVel++;
    }
}

void KinematicUnitDynamicSimulation::setJointTorques(const NameValueMap& targetJointTorques, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    // store current joint torques
    NameValueMap::const_iterator itAng = targetJointTorques.begin();

    while (itAng != targetJointTorques.end())
    {
        currentJointTorqueTargets[itAng->first] = itAng->second;
        itAng++;
    }
}

void KinematicUnitDynamicSimulation::setJointAccelerations(const NameValueMap& targetJointAccelerations, const Ice::Current& c)
{
}

void KinematicUnitDynamicSimulation::setJointDecelerations(const NameValueMap& targetJointDecelerations, const Ice::Current& c)
{
}

void KinematicUnitDynamicSimulation::step(double stepSizeMs, const Ice::Current& c)
{
    this->stepSizeMs = stepSizeMs;
    periodicExec();
}

void KinematicUnitDynamicSimulation::requestJoints(const Ice::StringSeq& joints, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    // if one of the joints belongs to this unit, lock access to this unit for other components except for the requesting one
    for (Ice::StringSeq::const_iterator it = joints.begin(); it != joints.end(); it++)
    {
        if (nodeNames.find(*it) != nodeNames.end())
        {
            KinematicUnit::request(c);
            return;
        }
    }
}

void KinematicUnitDynamicSimulation::releaseJoints(const Ice::StringSeq& joints, const Ice::Current& c)
{
    boost::recursive_mutex::scoped_lock scoped_lock(accessMutex);

    // if one of the joints belongs to this unit, unlock access
    for (Ice::StringSeq::const_iterator it = joints.begin(); it != joints.end(); it++)
    {
        if (nodeNames.find(*it) != nodeNames.end())
        {
            KinematicUnit::release(c);
            return;
        }
    }

}


void KinematicUnitDynamicSimulation::periodicExec()
{
    if (!simulatorPrx)
    {
        return;
    }

    float loopTime = 0;
    if (continuousMode)
    {
        loopTime = (float)execTask->getInterval() / 1000.0f; // in s
    }
    else
    {
        loopTime = (double)stepSizeMs / 1000.0f; //in s
    }
    float factorLoopTime = 1.1f;

    {
        boost::recursive_mutex::scoped_lock lock(accessMutex);

        NameValueMap applyPosMap;
        NameValueMap applyVelMap;
        NameValueMap applyPosVelMap_pos;
        NameValueMap applyPosVelMap_vel;
        NameValueMap applyTorMap;
        NameControlModeMap::const_iterator itMode = currentControlModes.begin();

        while (itMode != currentControlModes.end())
        {
            switch (itMode->second)
            {
                case ePositionControl:
                    if (currentJointTargets.find(itMode->first) != currentJointTargets.end())
                    {
                        applyPosMap[itMode->first] = currentJointTargets.find(itMode->first)->second;
                    }

                    break;

                case eVelocityControl:
                {
                    auto it = currentJointVelTargets.find(itMode->first);
                    if (it != currentJointVelTargets.end())
                    {
                        if (mapVelToPosVel)
                        {
                            // compute pos
                            if (currentJointValues.find(itMode->first) != currentJointValues.end())
                            {
                                float posTarget = currentJointValues[itMode->first] + it->second * loopTime * factorLoopTime;
                                applyPosVelMap_pos[itMode->first] = posTarget;
                                applyPosVelMap_vel[itMode->first] = it->second;
                            }
                        }
                        else
                        {
                            // just use velocity
                            applyVelMap[itMode->first] = currentJointVelTargets.find(itMode->first)->second;
                        }
                    }
                }

                break;

                case eTorqueControl:
                    if (currentJointTorqueTargets.find(itMode->first) != currentJointTorqueTargets.end())
                    {
                        applyTorMap[itMode->first] = currentJointTorqueTargets.find(itMode->first)->second;
                    }

                    break;

                case ePositionVelocityControl:
                    if (currentJointVelTargets.find(itMode->first) != currentJointVelTargets.end() &&
                        currentJointTargets.find(itMode->first) != currentJointTargets.end())
                    {
                        applyPosVelMap_pos[itMode->first] = currentJointTargets.find(itMode->first)->second;
                        applyPosVelMap_vel[itMode->first] = currentJointVelTargets.find(itMode->first)->second;
                    }

                    break;

                case eUnknown:
                case eDisabled:
                default:
                    // do nothing
                    ;
            }

            itMode++;
        }
        auto batchPrx = simulatorPrx->ice_batchOneway();
        if (applyPosMap.size() > 0)
        {
            batchPrx->actuateRobotJointsPos(robotName, applyPosMap);
        }

        if (applyVelMap.size() > 0)
        {
            batchPrx->actuateRobotJointsVel(robotName, applyVelMap);
        }

        if (applyTorMap.size() > 0)
        {
            batchPrx->actuateRobotJointsTorque(robotName, applyTorMap);
        }

        if (applyPosVelMap_pos.size() > 0)
        {
            batchPrx->actuateRobotJoints(robotName, applyPosVelMap_pos, applyPosVelMap_vel);
        }
        batchPrx->ice_flushBatchRequests();

        // command only once
        if (!mapVelToPosVel) // dont clear this in pos-velo mode because it is needed to calculate next position target
        {
            currentJointVelTargets.clear();
        }
        currentJointTargets.clear();
        currentJointTorqueTargets.clear();
    }
}




