/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::LaserScannerSimulation
 * @author     Fabian Paus ( fabian dot paus at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "LaserScannerSimulation.h"

#include <ArmarXCore/observers/variant/TimestampVariant.h>

#include <VirtualRobot/SceneObjectSet.h>
#include <VirtualRobot/Visualization/VisualizationFactory.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>

#include <chrono>

typedef std::chrono::high_resolution_clock Clock;
typedef Clock::duration Duration;

using namespace armarx;

static std::size_t nextMultipleOf(std::size_t value, std::size_t multiple)
{
    if (multiple == 0)
    {
        return value;
    }

    std::size_t remainder = value % multiple;
    if (remainder == 0)
    {
        return value;
    }

    return value + multiple - remainder;
}

static std::vector<std::string> splitProperty(Property<std::string> prop)
{
    std::vector<std::string> result;
    std::string propString(prop.getValue());
    boost::split(result, propString, boost::is_any_of(","));
    return result;
}

struct GridDimension
{
    float originX;
    float originY;
    std::size_t sizeX;
    std::size_t sizeY;
};

GridDimension calculateGridDimension(VirtualRobot::SceneObjectSetPtr const& objects, float gridStepSize)
{
    GridDimension result;
    result.originX = 0.0f;
    result.originY = 0.0f;
    result.sizeX = 1;
    result.sizeY = 1;
    if (objects->getSize() == 0)
    {
        return result;
    }

    // Determine the size the of the grid using the objects
    float minX = FLT_MAX;
    float minY = FLT_MAX;
    float maxX = -FLT_MAX;
    float maxY = -FLT_MAX;
    for (unsigned int objIndex = 0; objIndex < objects->getSize(); ++objIndex)
    {
        VirtualRobot::SceneObjectPtr object = objects->getSceneObject(objIndex);
        VirtualRobot::BoundingBox boundingBox = object->getCollisionModel()->getBoundingBox();
        Eigen::Vector3f minPos = boundingBox.getMin();
        Eigen::Vector3f maxPos = boundingBox.getMax();
        if (maxPos.x() > maxX)
        {
            maxX = maxPos.x();
        }
        if (minPos.x() < minX)
        {
            minX = minPos.x();
        }
        if (maxPos.y() > maxY)
        {
            maxY = maxPos.y();
        }
        if (minPos.y() < minY)
        {
            minY = minPos.y();
        }
    }

    float gridLengthX = maxX - minX;
    float gridLengthY = maxY - minY;
    int gridSizeX = (int)std::ceil(gridLengthX / gridStepSize);
    gridSizeX = nextMultipleOf(gridSizeX, sizeof(std::size_t) * CHAR_BIT);
    int gridSizeY = (int)std::ceil(gridLengthY / gridStepSize);

    result.originX = minX;
    result.originY = minY;
    result.sizeX = std::max(gridSizeX, 1);
    result.sizeY = std::max(gridSizeY, 1);
    return result;
}

void fillOccupancyGrid(OccupancyGrid* grid, float gridCellSize, std::vector<VirtualRobot::SceneObjectPtr> const& sceneObjects)
{
    // Get all the objects from the working memory
    VirtualRobot::SceneObjectSetPtr objects(new VirtualRobot::SceneObjectSet());
    //ARMARX_INFO_S << "Object count: " << objects->getSize();
    for (auto& object : sceneObjects)
    {
        //ARMARX_INFO_S << "Object: " << object->getName();
        objects->addSceneObject(object);
    }

    // Create a occupancy grid
    float gridStepSize = gridCellSize;
    GridDimension gridDim = calculateGridDimension(objects, gridStepSize);
    ARMARX_INFO_S << "Creating grid with size (" << gridDim.sizeX << ", " << gridDim.sizeY << ")";
    float boxPosZ = 500.0f;
    grid->init(gridDim.sizeX, gridDim.sizeY, gridDim.originX, gridDim.originY, gridStepSize);
    // - Sample the space with a box (10cm x 10 cm x 10 cm)

    VirtualRobot::CollisionCheckerPtr collisionChecker = VirtualRobot::CollisionChecker::getGlobalCollisionChecker();
    VirtualRobot::VisualizationFactoryPtr factory(new VirtualRobot::CoinVisualizationFactory());
    float boxSize = 1.1f * gridStepSize;
    VirtualRobot::VisualizationNodePtr boxVisu = factory->createBox(boxSize, boxSize, boxSize, 0.0f, 1.0f, 0.0f);
    VirtualRobot::CollisionModelPtr boxCollisionModel(new VirtualRobot::CollisionModel(boxVisu, "", collisionChecker));
    VirtualRobot::SceneObjectPtr box(new VirtualRobot::SceneObject("my_box", boxVisu, boxCollisionModel));

    if (!collisionChecker)
    {
        ARMARX_WARNING_S << "No global collision checker found";
        return;
    }

    ARMARX_INFO_S << "Filling occupancy grid";

    for (std::size_t indexY = 0; indexY < grid->sizeY; ++indexY)
    {
        for (std::size_t indexX = 0; indexX < grid->sizeX; ++indexX)
        {
            if (indexY == 0 || indexX == 0 || indexY == grid->sizeY - 1 || indexX == grid->sizeX - 1)
            {
                grid->setOccupied(indexX, indexY, true);
            }
            else
            {
                Eigen::Vector2f pos = grid->getCentralPosition(indexX, indexY);
                Eigen::Vector3f boxPos(pos.x(), pos.y(), boxPosZ);
                Eigen::Matrix4f boxPose = Eigen::Affine3f(Eigen::Translation3f(boxPos)).matrix();
                box->setGlobalPose(boxPose);
                // - Check collisions with the other objects
                bool collision = collisionChecker->checkCollision(box, objects);
                // - Mark each field as occupied or free
                grid->setOccupied(indexX, indexY, collision);
            }
        }
    }
}

void LaserScannerSimulation::onInitComponent()
{
    scanners.clear();

    // Necessary to initialize SoDB and Qt, SoQt::init (otherwise the application crashes at startup)
    VirtualRobot::init("SimulatorViewerApp");
    worldVisu = Component::create<ArmarXPhysicsWorldVisualization>(getIceProperties(), getName() + "_PhysicsWorldVisualization");
    getArmarXManager()->addObject(worldVisu);

    topicName = getProperty<std::string>("LaserScannerTopicName").getValue();
    ARMARX_INFO << "Reporting on topic \"" << topicName << "\"";
    offeringTopic(topicName);

    robotStateName = getProperty<std::string>("RobotStateComponentName").getValue();
    ARMARX_INFO << "Using RobotStateComponent \"" << robotStateName << "\"";
    usingProxy(robotStateName);

    debugDrawerName = getProperty<std::string>("DebugDrawerTopicName").getValue();
    ARMARX_INFO << "Using DebugDrawerTopic \"" << debugDrawerName << "\"";
    // No usingProxy for the DebugDrawerTopic? or is it a topic?

    topicReplayerDummy = getProperty<bool>("TopicReplayerDummy").getValue();
    updatePeriod = getProperty<int>("UpdatePeriod").getValue();
    gridCellSize = getProperty<float>("GridCellSize").getValue();

    auto framesStrings = splitProperty(getProperty<std::string>("Frames"));
    auto deviceStrings = splitProperty(getProperty<std::string>("Devices"));
    auto minAnglesStrings = splitProperty(getProperty<std::string>("MinAngles"));
    auto maxAnglesStrings = splitProperty(getProperty<std::string>("MaxAngles"));
    auto stepsStrings = splitProperty(getProperty<std::string>("Steps"));
    auto noiseStrings = splitProperty(getProperty<std::string>("NoiseStdDev"));
    std::size_t scannerSize = framesStrings.size();
    if (deviceStrings.size() == 1)
    {
        deviceStrings.resize(scannerSize, deviceStrings.front());
    }
    else if (deviceStrings.size() != scannerSize)
    {
        ARMARX_WARNING << "Unexpected size of property Devices (expected " << scannerSize << " but got " << deviceStrings.size() << ")";
        return;
    }
    if (minAnglesStrings.size() == 1)
    {
        minAnglesStrings.resize(scannerSize, minAnglesStrings.front());
    }
    else if (minAnglesStrings.size() != scannerSize)
    {
        ARMARX_WARNING << "Unexpected size of property MinAngles (expected " << scannerSize << " but got " << minAnglesStrings.size() << ")";
        return;
    }
    if (maxAnglesStrings.size() == 1)
    {
        maxAnglesStrings.resize(scannerSize, maxAnglesStrings.front());
    }
    else if (maxAnglesStrings.size() != scannerSize)
    {
        ARMARX_WARNING << "Unexpected size of property MaxAngles (expected " << scannerSize << " but got " << maxAnglesStrings.size() << ")";
        return;
    }
    if (stepsStrings.size() == 1)
    {
        stepsStrings.resize(scannerSize, stepsStrings.front());
    }
    else if (stepsStrings.size() != scannerSize)
    {
        ARMARX_WARNING << "Unexpected size of property Steps (expected " << scannerSize << " but got " << stepsStrings.size() << ")";
        return;
    }
    if (noiseStrings.size() == 1)
    {
        noiseStrings.resize(scannerSize, noiseStrings.front());
    }
    else if (noiseStrings.size() != scannerSize)
    {
        ARMARX_WARNING << "Unexpected size of property NoiseStdDev (expected " << scannerSize << " but got " << noiseStrings.size() << ")";
        return;
    }

    scanners.reserve(scannerSize);
    connectedDevices.clear();
    for (std::size_t i = 0; i < scannerSize; ++i)
    {
        LaserScannerSimUnit scanner;
        scanner.frame = framesStrings[i];
        try
        {
            scanner.minAngle = boost::lexical_cast<float>(minAnglesStrings[i]);
            scanner.maxAngle = boost::lexical_cast<float>(maxAnglesStrings[i]);
            scanner.noiseStdDev = boost::lexical_cast<float>(noiseStrings[i]);
            scanner.steps = std::stoi(stepsStrings[i]);
        }
        catch (std::exception const& ex)
        {
            ARMARX_INFO << "Scanner[" << i << "] Config error: " << ex.what();
            continue;
        }

        scanners.push_back(scanner);

        LaserScannerInfo info;
        info.device = topicReplayerDummy ? deviceStrings[i] : scanner.frame;
        info.frame = scanner.frame;
        info.minAngle = scanner.minAngle;
        info.maxAngle = scanner.maxAngle;
        info.stepSize = (info.maxAngle - info.minAngle) / scanner.steps;
        connectedDevices.push_back(info);

        ARMARX_INFO << "Scanner[" << i << "]: " << scanner.frame << ", "
                    << scanner.minAngle << ", " << scanner.maxAngle << ", " << scanner.steps;
    }
}

void LaserScannerSimulation::onConnectComponent()
{
    if (topicReplayerDummy)
    {
        ARMARX_INFO << "Fake connect (component is used for topic replay)";
        return;
    }
    topic = getTopic<LaserScannerUnitListenerPrx>(topicName);
    robotState = getProxy<RobotStateComponentInterfacePrx>(robotStateName);
    sharedRobot = robotState->getSynchronizedRobot();
    debugDrawer = getTopic<DebugDrawerInterfacePrx>(debugDrawerName);

    for (LaserScannerSimUnit& scanner : scanners)
    {
        try
        {
            scanner.frameNode = sharedRobot->getRobotNode(scanner.frame);
        }
        catch (std::exception const& ex)
        {
            ARMARX_WARNING << "Error while querying robot frame: " << scanner.frame << " " << ex.what();
            scanner.frameNode = nullptr;
        }

        if (!scanner.frameNode)
        {
            ARMARX_WARNING << "Tried to use a non-existing robot node as frame for the laser scanner simulation: " << scanner.frame;
        }
    }

    if (task)
    {
        task->stop();
    }

    // Don't forget to sync the data otherwise there may be no objects
    std::vector<VirtualRobot::SceneObjectPtr> objects;
    while (objects.empty())
    {
        worldVisu->synchronizeVisualizationData();
        objects = worldVisu->getObjects();
        ARMARX_INFO << deactivateSpam(5) << "Got " << objects.size() << " objects from the simulator";
        if (objects.empty())
        {
            ARMARX_INFO << deactivateSpam() << "Could not get any objects from the simulator after syncing";
            TimeUtil::MSSleep(100);
        }
    }
    fillOccupancyGrid(&grid, gridCellSize, objects);

    task = new PeriodicTask<LaserScannerSimulation>(this, &LaserScannerSimulation::updateScanData, updatePeriod, false, "LaserScannerSimUpdate");
    task->start();
}


void LaserScannerSimulation::onDisconnectComponent()
{
    if (task)
    {
        task->stop();
        task = nullptr;
    }
}


void LaserScannerSimulation::onExitComponent()
{
    if (worldVisu)
    {
        getArmarXManager()->removeObjectBlocking(worldVisu);
        worldVisu = nullptr;
    }
}

armarx::PropertyDefinitionsPtr LaserScannerSimulation::createPropertyDefinitions()
{
    return armarx::PropertyDefinitionsPtr(new LaserScannerSimulationPropertyDefinitions(
            getConfigIdentifier()));
}

std::string LaserScannerSimulation::getReportTopicName(const Ice::Current&) const
{
    return topicName;
}

LaserScannerInfoSeq LaserScannerSimulation::getConnectedDevices(const Ice::Current&) const
{
    return connectedDevices;
}

void LaserScannerSimulation::updateScanData()
{
    auto startTime = Clock::now();
    TimestampVariantPtr now(new TimestampVariant(TimeUtil::GetTime()));

    for (LaserScannerSimUnit const& scanner : scanners)
    {
        if (!scanner.frameNode)
        {
            continue;
        }
        PosePtr scannerPoseP = PosePtr::dynamicCast(scanner.frameNode->getGlobalPose());
        Eigen::Matrix4f scannerPose = scannerPoseP->toEigen();
        Eigen::Vector2f position = scannerPose.col(3).head<2>();

        Eigen::Matrix3f scannerRot = scannerPose.block<3, 3>(0, 0);
        Eigen::Vector2f yWorld(0.0f, 1.0f);
        Eigen::Vector2f yScanner = scannerRot.col(1).head<2>();
        float theta = acos(yWorld.dot(yScanner));
        if (yScanner.x() >= 0.0f)
        {
            theta = -theta;
        }

        float minAngle = scanner.minAngle;
        float maxAngle = scanner.maxAngle;
        int scanSteps = scanner.steps;

        LaserScan scan;
        scan.reserve(scanSteps);

        std::normal_distribution<float> dist(0.0f, scanner.noiseStdDev);
        for (int i = 0; i < scanSteps; ++i)
        {
            LaserScanStep step;
            step.angle = minAngle + i * (maxAngle - minAngle) / (scanSteps - 1);
            step.distance = 1000.0f;

            Eigen::Vector2f scanDirGlobal = Eigen::Rotation2Df(step.angle + theta) * Eigen::Vector2f(0.0f, 1.0f);
            float distance = grid.computeDistance(position, scanDirGlobal);
            if (distance > 0.0f)
            {
                step.distance = distance + dist(engine);
                scan.push_back(step);
            }
        }

        topic->reportSensorValues(scanner.frame, scanner.frame, scan, now);
    }

    auto endTime = Clock::now();
    auto timeElapsed = endTime - startTime;
    Duration::rep timeElapsedMicroSecs = std::chrono::duration_cast<std::chrono::microseconds>(timeElapsed).count();
    ARMARX_INFO << deactivateSpam(10) << "Time to simulate laser scanners: " << timeElapsedMicroSecs * 0.001f << " ms";
}

