/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::LaserScannerSimulation
 * @author     Fabian Paus ( fabian dot paus at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_ArmarXSimulation_LaserScannerSimulation_H
#define _ARMARX_COMPONENT_ArmarXSimulation_LaserScannerSimulation_H

#include <ArmarXSimulation/components/ArmarXPhysicsWorldVisualization/ArmarXPhysicsWorldVisualization.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <RobotAPI/interface/core/RobotState.h>
#include <RobotAPI/components/units/SensorActorUnit.h>
#include <RobotAPI/interface/units/LaserScannerUnit.h>
#include <RobotAPI/interface/visualization/DebugDrawerInterface.h>

#include <VirtualRobot/VirtualRobot.h>
#include <boost/dynamic_bitset.hpp>
#include <Eigen/Eigen>
#include <random>

namespace armarx
{

    struct OccupancyGrid
    {
        void init(std::size_t gridSizeX, std::size_t gridSizeY, float originX, float originY, float gridStepSize)
        {
            sizeX = gridSizeX;
            sizeY = gridSizeY;
            this->originX = originX;
            this->originY = originY;
            this->stepSize = gridStepSize;
            std::size_t gridSize = sizeX * sizeY;
            occupied.clear();
            occupied.resize(gridSize);
        }

        bool isOccupied(std::size_t indexX, std::size_t indexY) const
        {
            std::size_t index = indexY * sizeX + indexX;
            if (index < occupied.size())
            {
                return occupied[index];
            }
            return false;
        }

        void setOccupied(std::size_t indexX, std::size_t indexY, bool posOccupied)
        {
            std::size_t index = indexY * sizeX + indexX;
            if (index < occupied.size())
            {
                occupied[index] = posOccupied;
            }
        }

        Eigen::Vector2f getCentralPosition(std::size_t indexX, std::size_t indexY) const
        {
            float posX = originX + indexX * stepSize;
            float posY = originY + indexY * stepSize;
            return Eigen::Vector2f(posX, posY);
        }

        bool isOccupied(float posX, float posY) const
        {
            std::size_t indexX = (std::size_t)std::roundf((posX - originX) / stepSize);
            std::size_t indexY = (std::size_t)std::roundf((posY - originY) / stepSize);
            return isOccupied(indexX, indexY);
        }

        float computeDistance(Eigen::Vector2f pos, Eigen::Vector2f dir)
        {
            std::size_t indexX = (std::size_t)std::roundf((pos.x() - originX) / stepSize);
            std::size_t indexY = (std::size_t)std::roundf((pos.y() - originY) / stepSize);
            if (indexX >= sizeX || indexY >= sizeY)
            {
                return 0.0f;
            }

            // We are inside the grid
            float gridMaxX = originX + stepSize * sizeX;
            float gridMaxY = originY + stepSize * sizeY;
            float distanceX = 0.0f;
            if (dir.x() > 0.0f)
            {
                distanceX = std::abs((gridMaxX - pos.x()) / dir.x());
            }
            else if (dir.x() < 0.0f)
            {
                distanceX = std::abs((pos.x() - originX) / dir.x());
            }
            float distanceY = 0.0f;
            if (dir.y() > 0.0f)
            {
                distanceY = std::abs((gridMaxY - pos.y()) / dir.y());
            }
            else if (dir.y() < 0.0f)
            {
                distanceY = std::abs((pos.y() - originY) / dir.y());
            }

            // What is the max distance we can travel along dir until we leave the grid?
            float maxDistance = std::min(distanceX, distanceY);
            float testStepSize = 0.5f * stepSize;
            Eigen::Vector2f testStep = testStepSize * dir;
            std::size_t lineSteps = std::ceil(maxDistance / testStepSize);
            for (std::size_t i = 0; i < lineSteps; ++i)
            {
                Eigen::Vector2f testPos = pos + i * testStep;
                if (isOccupied(testPos.x(), testPos.y()))
                {
                    return i * testStepSize;
                }
            }
            // TODO: Define the max distance which is returned if no occupied field was found
            return 0.0f;
        }

        std::size_t sizeX = 0;
        std::size_t sizeY = 0;
        float originX = 0.0f;
        float originY = 0.0f;
        float stepSize = 0.0f;
        boost::dynamic_bitset<std::size_t> occupied;
    };


    /**
     * @class LaserScannerSimulationPropertyDefinitions
     * @brief
     */
    class LaserScannerSimulationPropertyDefinitions:
        public armarx::ComponentPropertyDefinitions
    {
    public:
        LaserScannerSimulationPropertyDefinitions(std::string prefix):
            armarx::ComponentPropertyDefinitions(prefix)
        {
            defineOptionalProperty<std::string>("LaserScannerTopicName", "LaserScans", "Name of the laser scan topic.");
            defineOptionalProperty<std::string>("RobotStateComponentName", "RobotStateComponent", "Name of the RobotStateComponent to use.");
            defineOptionalProperty<std::string>("DebugDrawerTopicName", "DebugDrawerUpdates", "Visualize the results here.");
            defineOptionalProperty<std::string>("ReportVisuTopicName", "SimulatorVisuUpdates", "The topic on which the visualization updates are published.");

            defineOptionalProperty<bool>("TopicReplayerDummy", false, "Enable to serve the purpose of a topic replayer dummy");
            defineOptionalProperty<int>("UpdatePeriod", 25, "Update period for laser scans in ms");
            defineOptionalProperty<float>("GridCellSize", 20.0f, "Size of the grid cells used to generate the occupancy map in mm");
            defineOptionalProperty<std::string>("Frames", "LaserScanner_1", "Name of the frames to attach the sensor to (e.g. Node1,Node2,Node3)");
            defineOptionalProperty<std::string>("Devices", "127.0.0.1", "Name of the devices to simulate (e.g. Device1,Device2,Device3)");
            defineOptionalProperty<std::string>("MinAngles", "-2.3561945", "Minumum angles to be reported in rad (e.g -2.35,-1.27,0)");
            defineOptionalProperty<std::string>("MaxAngles", "+2.3561945", "Maxiumum angles to be reported in rad (e.g 2.35,1.27,3.14)");
            defineOptionalProperty<std::string>("Steps", "1081", "Number of single steps (angle, distance) per scan (e.g 1081,360,270)");
            defineOptionalProperty<std::string>("NoiseStdDev", "40.0", "Noise is added to the distance of single steps (e.g 40,30,20)");
        }
    };

    struct LaserScannerSimUnit
    {
        std::string frame;
        float minAngle;
        float maxAngle;
        float noiseStdDev;
        int steps;
        SharedRobotNodeInterfacePrx frameNode;
    };

    /**
     * @defgroup Component-LaserScannerSimulation LaserScannerSimulation
     * @ingroup ArmarXSimulation-Components
     * A description of the component LaserScannerSimulation.
     *
     * @class LaserScannerSimulation
     * @ingroup Component-LaserScannerSimulation
     * @brief Brief description of class LaserScannerSimulation.
     *
     * Detailed description of class LaserScannerSimulation.
     */
    class LaserScannerSimulation :
        virtual public armarx::SensorActorUnit,
        virtual public armarx::LaserScannerUnitInterface
    {
    public:
        /**
         * @see armarx::ManagedIceObject::getDefaultName()
         */
        virtual std::string getDefaultName() const
        {
            return "LaserScannerSimulation";
        }

    protected:
        /**
         * @see armarx::ManagedIceObject::onInitComponent()
         */
        virtual void onInitComponent();

        /**
         * @see armarx::ManagedIceObject::onConnectComponent()
         */
        virtual void onConnectComponent();

        /**
         * @see armarx::ManagedIceObject::onDisconnectComponent()
         */
        virtual void onDisconnectComponent();

        /**
         * @see armarx::ManagedIceObject::onExitComponent()
         */
        virtual void onExitComponent();

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        virtual armarx::PropertyDefinitionsPtr createPropertyDefinitions();

        std::string getReportTopicName(const Ice::Current&) const override;

        LaserScannerInfoSeq getConnectedDevices(const Ice::Current&) const override;

    private:
        void updateScanData();

    private:
        std::string topicName;
        LaserScannerUnitListenerPrx topic;
        std::string robotStateName;
        armarx::RobotStateComponentInterfacePrx robotState;
        armarx::SharedRobotInterfacePrx sharedRobot;
        std::string debugDrawerName;
        armarx::DebugDrawerInterfacePrx debugDrawer;

        ArmarXPhysicsWorldVisualizationPtr worldVisu;

        bool topicReplayerDummy = false;
        int updatePeriod = 25;
        float gridCellSize = 20.0f;
        std::vector<LaserScannerSimUnit> scanners;
        PeriodicTask<LaserScannerSimulation>::pointer_type task;

        OccupancyGrid grid;
        std::mt19937 engine;

        LaserScannerInfoSeq connectedDevices;
    };
}

#endif
