/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::units
 * @author     Manfred Kroehnert (manfred dot kroehnert at kit dot edu)
 * @date       2013
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_PLATFORM_UNIT_DYNAMIC_SIMULATION_H
#define _ARMARX_COMPONENT_PLATFORM_UNIT_DYNAMIC_SIMULATION_H

#include <RobotAPI/components/units/PlatformUnit.h>

#include <RobotAPI/libraries/core/Pose.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <ArmarXCore/core/system/ImportExportComponent.h>

#include <ArmarXSimulation/interface/simulator/SimulatorInterface.h>

#include <IceUtil/Time.h>
#include <MemoryX/interface/component/WorkingMemoryInterface.h>

#include <string>
#include <cmath>

namespace armarx
{
    /**
     * @class PlatformUnitDynamicSimulationPropertyDefinitions
     * @brief
     */
    class PlatformUnitDynamicSimulationPropertyDefinitions :
        public PlatformUnitPropertyDefinitions
    {
    public:
        PlatformUnitDynamicSimulationPropertyDefinitions(std::string prefix) :
            PlatformUnitPropertyDefinitions(prefix)
        {
            defineOptionalProperty<int>("IntervalMs", 100, "The time in milliseconds between two calls to the simulation method.");

            //assures that the platform still gets some amount of velocity when near target
            defineOptionalProperty<float>("LinearVelocityMin", 0.5, "Minimum Linear velocity of the platform (unit: m/sec).");
            defineOptionalProperty<float>("LinearVelocityMax", 1.5, "Maximum Linear velocity of the platform (unit: m/sec).");
            defineOptionalProperty<float>("LinearAccelerationMax", 0.9, "Maximum Linear Acceleration (unit: m/sec^2");
            defineOptionalProperty<float>("AngularVelocityMax", 0.8, "Maximum Angular velocity of the platform (unit: rad/sec).");
            defineOptionalProperty<float>("AngularAccelerationMax", 1.9, "Maximum Angular Acceleration (unit: rad/sec^2");
            defineOptionalProperty<float>("PlatformOrientationOffset", 0, "Rotation offset of the platform (unit: rad).");
            defineOptionalProperty<std::string>("RobotName", "Armar3", "Name of the Robot to use.");
            defineOptionalProperty<std::string>("PlatformName", "Platform", "Name of the Platform node to use.");
            defineOptionalProperty<std::string>("SimulatorProxyName", "Simulator", "Name of the simulator proxy to use.");
            defineOptionalProperty<std::string>("WorkingMemoryName", "WorkingMemory", "Name of the WorkingMemory that should be used");
        }
    };

    /**
     * @class PlatformUnitDynamicSimulation
     * @brief This unit connects to the physics simulator topic (default: "Simulator") and handles platform movements.
     *
     * @ingroup SensorActorUnits
     * @ingroup ArmarXSimulatorComponents
     */
    class PlatformUnitDynamicSimulation :
        virtual public PlatformUnit,
        virtual public PlatformUnitDynamicSimulationInterface
    {
    public:

        // inherited from Component
        virtual std::string getDefaultName() const
        {
            return "PlatformUnit";
        }

        virtual void onInitPlatformUnit();
        virtual void onStartPlatformUnit();
        virtual void onStopPlatformUnit();
        virtual void onExitPlatformUnit();

        void simulationFunction();

        // proxy implementation
        void moveTo(Ice::Float targetPlatformPositionX, Ice::Float targetPlatformPositionY, Ice::Float targetPlatformRotation,  Ice::Float positionalAccuracy, Ice::Float orientationalAccuracy, const Ice::Current& c = ::Ice::Current());
        void move(float targetPlatformVelocityX, float targetPlatformVelocityY, float targetPlatformVelocityRotation, const Ice::Current& c = Ice::Current());
        void moveRelative(float targetPlatformOffsetX, float targetPlatformOffsetY, float targetPlatformOffsetRotation, float positionalAccuracy, float orientationalAccuracy, const Ice::Current& c = Ice::Current());
        void setMaxVelocities(float linearVelocity, float angularVelocity, const Ice::Current& c = Ice::Current());

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

    protected:
        void updateCurrentPose(const PoseBasePtr& newPose);
        void updateCurrentVelocity(const Vector3BasePtr& translationVel, const Vector3BasePtr& rotationVel);

        // implement SimulatorRobotListener to receive robot updates from simulator
        void reportJointAngles(const NameValueMap& actualJointAngles, Ice::Long timestamp, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointVelocities(const NameValueMap& actualJointVelocities, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointTorques(const NameValueMap& actualJointTorques, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportPose(const PoseBasePtr& actualPose, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportVelocity(const Vector3BasePtr& translationVel, const Vector3BasePtr& rotationVel, bool aValueChanged, const Ice::Current& c = ::Ice::Current());

        //calc acceleration based on distance and current speed
        Eigen::Vector3f calculateLinearVelocity(Eigen::Vector3f distance, Eigen::Vector3f curV);
        float calculateAngularVelocity(float source, float target, float curV);

        float clampf(float x, float min, float max)
        {
            return std::max<float>(std::min<float>(x, max), min);
        }

        boost::mutex currentPoseMutex;
        IceUtil::Time lastExecutionTime;
        int intervalMs;

        Vector3 currentPosition;
        Vector3 targetPosition;
        ::Ice::Float currentRotation;

        ::Ice::Float targetRotation;

        Vector3 currentVelocity;
        ::Ice::Float currentRotationVelocity;

        ::Ice::Float linearVelocityMin;
        ::Ice::Float linearVelocityMax;
        ::Ice::Float linearAccelerationMax;
        ::Ice::Float platformOrientationOffset;

        ::Ice::Float angularVelocityMax;
        ::Ice::Float angularAccelerationMax;

        ::Ice::Float positionalAccuracy;
        ::Ice::Float orientationalAccuracy;

        std::string referenceFrame;
        std::string simulatorPrxName;
        std::string robotName;
        std::string platformName;

        ControlMode platformControlMode;


        SimulatorInterfacePrx simulatorPrx;

        PeriodicTask<PlatformUnitDynamicSimulation>::pointer_type simulationTask;

        // PlatformUnitInterface interface
    public:
        void stopPlatform(const Ice::Current& current);
    private:

        memoryx::WorkingMemoryInterfacePrx memoryPrx;
    };
}

#endif
