/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation
 * @author     Nikolaus Vahrenkamp
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SelfLocalizationDynamicSimulation.h"

// Core
#include <ArmarXCore/core/system/ArmarXDataPath.h>

// Robot API
#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

// MemoryX
#include <MemoryX/core/entity/Entity.h>
#include <MemoryX/core/entity/EntityRef.h>
#include <MemoryX/core/MemoryXCoreObjectFactories.h>
#include <IceUtil/UUID.h>

using namespace armarx;
using namespace memoryx;
using namespace memoryx::EntityWrappers;


SelfLocalizationDynamicSimulation::SelfLocalizationDynamicSimulation()
{
    robotAgentId = IceUtil::generateUUID();
}


void SelfLocalizationDynamicSimulation::onInitComponent()
{
    ARMARX_INFO << "Init...";
    usingProxy(getProperty<std::string>("RobotStateComponentName").getValue());

    std::string simPrxName = getProperty<std::string>("SimulatorName").getValue();

    if (!simPrxName.empty())
    {
        usingProxy(simPrxName);
    }

    usingProxy(getProperty<std::string>("WorkingMemoryName").getValue());
}


void SelfLocalizationDynamicSimulation::onConnectComponent()
{
    ARMARX_INFO << "Connecting...";
    std::string simPrxName = getProperty<std::string>("SimulatorName").getValue();

    if (!simPrxName.empty())
    {
        simulatorProxy = getProxy<SimulatorInterfacePrx>(simPrxName);
    }

    ARMARX_CHECK_EXPRESSION(simulatorProxy);
    ARMARX_INFO << "Retrieved simulator proxy...";

    robotStateComponentPrx = getProxy<RobotStateComponentInterfacePrx>(getProperty<std::string>("RobotStateComponentName").getValue());
    ARMARX_CHECK_EXPRESSION(robotStateComponentPrx);
    ARMARX_INFO << "Retrieved robot state proxy...";

    memoryPrx = getProxy<memoryx::WorkingMemoryInterfacePrx>(getProperty<std::string>("WorkingMemoryName").getValue());
    ARMARX_CHECK_EXPRESSION(memoryPrx);
    ARMARX_INFO << "Retrieved WM proxy...";

    agentsMemoryPrx = memoryPrx->getAgentInstancesSegment();
    ARMARX_CHECK_EXPRESSION(agentsMemoryPrx);
    ARMARX_INFO << "Retrieved WM agent segment ...";



    robotName = getProperty<std::string>("RobotName").getValue();
    ARMARX_CHECK_EXPRESSION_W_HINT(simulatorProxy->hasRobot(robotName), "Robot with given name not available in armarx simulator: " + robotName);

    std::string agentName = getProperty<std::string>("AgentName").isSet() ? getProperty<std::string>("AgentName").getValue() : robotName;

    ARMARX_INFO << "Creating robot agent with name " << agentName;
    robotAgent = new AgentInstance(agentName);
    auto robot = robotStateComponentPrx->getSynchronizedRobot();
    robotAgent->setSharedRobot(robot);
    robotAgent->setName(agentName);
    robotAgent->setAgentFilePath(robotStateComponentPrx->getRobotFilename());
    ARMARX_INFO << "Creating robot agent...done";


    if (getProperty<bool>("ContinuousMode"))
    {
        // periodic task setup
        cycleTime = 30;

        if (execTask)
        {
            execTask->stop();
        }

        execTask = new PeriodicTask<SelfLocalizationDynamicSimulation>(this, &SelfLocalizationDynamicSimulation::periodicExec, cycleTime, false, "SelfLocalizationDynamicSimulationCalculation");
        execTask->start();
        execTask->setDelayWarningTolerance(100);
    }
}


void SelfLocalizationDynamicSimulation::onDisconnectComponent()
{
    ARMARX_INFO << "Disconnect...";

    if (execTask)
    {
        execTask->stop();
    }

    robotAgentId.clear();
}

void SelfLocalizationDynamicSimulation::step(const Ice::Current& c)
{
    ARMARX_INFO << deactivateSpam(1) << "step!";
    reportRobotPose();
}



void SelfLocalizationDynamicSimulation::periodicExec()
{
    reportRobotPose();
}



void SelfLocalizationDynamicSimulation::reportRobotPose()
{
    try
    {
        ScopedLock lock(dataMutex);
        ARMARX_DEBUG << "Reporting robot agent:" << robotAgent->getName() << ", robotAgentId:" << robotAgentId << ", self:" << this->getName();

        PoseBasePtr poseBase = simulatorProxy->getRobotPose(robotName);
        Eigen::Matrix4f gp;

        if (poseBase)
        {
            gp = PosePtr::dynamicCast(poseBase)->toEigen();
        }
        else
        {
            ARMARX_WARNING << deactivateSpam(10) << "No robot pose from simulator...";
            gp = Eigen::Matrix4f::Identity();
        }

        Eigen::Matrix3f orientation = gp.block(0, 0, 3, 3);
        FramedOrientationPtr orientationPtr = new FramedOrientation(orientation, GlobalFrame, "");
        Eigen::Vector3f position = gp.block(0, 3, 3, 1);
        FramedPositionPtr positionPtr = new FramedPosition(position, GlobalFrame, "");
        robotAgent->setOrientation(orientationPtr);
        robotAgent->setPosition(positionPtr);

        robotAgentId = agentsMemoryPrx->upsertEntityByName(robotAgent->getName(), robotAgent);
        robotAgent->setId(robotAgentId);


    }
    catch (...)
    {
        ARMARX_WARNING << deactivateSpam(10) << "Could not update robot pose in agents segment...\n" << GetHandledExceptionString();
        robotAgentId.clear();
    }
}





