/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::Simulator
 * @author     Nikolaus Vahrenkamp ( vahrenkamp at kit dot edu )
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef __ARMARX_PHYSICS_WORLD__
#define __ARMARX_PHYSICS_WORLD__

#include <VirtualRobot/VirtualRobotCommon.h>
#include <SimDynamics/DynamicsEngine/DynamicsEngine.h>
#include <SimDynamics/DynamicsWorld.h>
#include <SimDynamics/DynamicsEngine/BulletEngine/BulletEngine.h>
#include <SimDynamics/DynamicsEngine/DynamicsRobot.h>
#include <SimDynamics/DynamicsEngine/DynamicsObject.h>
#include <SimDynamics/DynamicsEngine/BulletEngine/BulletRobotLogger.h>

#include <btBulletDynamicsCommon.h>
#include <LinearMath/btQuickprof.h>

#include <boost/thread/recursive_mutex.hpp>

#include <Eigen/Core>
#include <Eigen/Geometry>

// just needed for SceneData
#include <ArmarXSimulation/interface/simulator/SimulatorInterface.h>
#include <ArmarXCore/core/system/Synchronization.h>
#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>

#include "SimulatedWorld.h"

namespace armarx
{
    /*!
     * \brief The BulletPhysicsWorld class encapsulates the whole physics simulation and the corresponding data
     *
     */
    class BulletPhysicsWorld : public SimulatedWorld
    {
    public:

        BulletPhysicsWorld();
        ~BulletPhysicsWorld();


        virtual void initialize(int stepSizeMS, float bulletFixedTimeStepMS, int bulletFixedTimeStepMaxNrLoops, bool maxRealTimeSimSpeed = true, bool floorPlane = false, std::string floorTexture = std::string(), float floorFrictionCoeff = -1);

        virtual void actuateRobotJoints(const std::string& robotName, const std::map< std::string, float>& angles, const std::map< std::string, float>& velocities);
        virtual void actuateRobotJointsPos(const std::string& robotName, const std::map< std::string, float>& angles);
        virtual void actuateRobotJointsVel(const std::string& robotName, const std::map< std::string, float>& velocities);
        virtual void actuateRobotJointsTorque(const std::string& robotName, const std::map< std::string, float>& torques);
        virtual void setRobotPose(const std::string& robotName, const Eigen::Matrix4f& globalPose);

        virtual void applyForceRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& force);
        virtual void applyImpulseRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& impulse);
        virtual void applyTorqueRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& torque);

        virtual void applyForceObject(const std::string& objectName, const Eigen::Vector3f& force);
        virtual void applyTorqueObject(const std::string& objectName, const Eigen::Vector3f& torque);

        virtual bool hasObject(const std::string& instanceName);
        virtual void setObjectPose(const std::string& objectName, const Eigen::Matrix4f& globalPose);
        virtual void activateObject(const std::string& objectName);

        virtual bool hasRobot(const std::string& robotName);
        virtual bool hasRobotNode(const std::string& robotName, const std::string& robotNodeName);
        virtual float getRobotMass(const std::string& robotName);

        virtual std::map< std::string, float> getRobotJointAngles(const std::string& robotName);
        virtual float getRobotJointAngle(const std::string& robotName, const std::string& nodeName);
        virtual std::map< std::string, float> getRobotJointVelocities(const std::string& robotName);
        virtual float getRobotJointVelocity(const std::string& robotName, const std::string& nodeName);
        virtual std::map< std::string, float> getRobotJointTorques(const std::string& robotName);
        virtual ForceTorqueDataSeq getRobotForceTorqueSensors(const std::string& robotName);

        virtual float getRobotJointLimitLo(const std::string& robotName, const std::string& nodeName);
        virtual float getRobotJointLimitHi(const std::string& robotName, const std::string& nodeName);
        virtual Eigen::Matrix4f getRobotPose(const std::string& robotName);

        virtual float getRobotMaxTorque(const std::string& robotName, const std::string& nodeName);
        virtual void setRobotMaxTorque(const std::string& robotName, const std::string& nodeName, float maxTorque);

        virtual Eigen::Matrix4f getRobotNodePose(const std::string& robotName, const std::string& robotNodeName);

        virtual Eigen::Vector3f getRobotLinearVelocity(const std::string& robotName, const std::string& nodeName);
        virtual Eigen::Vector3f getRobotAngularVelocity(const std::string& robotName, const std::string& nodeName);

        virtual void setRobotLinearVelocity(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotAngularVelocity(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotLinearVelocityRobotRootFrame(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotAngularVelocityRobotRootFrame(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);

        virtual Eigen::Matrix4f getObjectPose(const std::string& objectName);


        virtual FramedPosePtr toFramedPose(const Eigen::Matrix4f& globalPose, const std::string& robotName, const std::string& frameName);

        /**
        * Perform one simulation step. Calculates the delta update time from the time that has passed since the last call.
        * This updates all models.
        */
        virtual void stepPhysicsRealTime();

        /**
        * Perform one simulation step.
        * This updates all models.
        */
        virtual void stepPhysicsFixedTimeStep();

        /*!
         * \brief
         * \return The number of steps * the timestep in MS
         */
        virtual int getFixedTimeStepMS();

        SimDynamics::BulletEnginePtr getEngine();

        std::vector<VirtualRobot::SceneObjectPtr> getObjects();

        std::vector<SimDynamics::DynamicsObjectPtr> getDynamicObjects();
        SimDynamics::DynamicsObjectPtr getObject(const std::string& objectName);


        SimDynamics::DynamicsObjectPtr getDynamicsObject(const std::string& objectName);

        virtual void enableLogging(const std::string& robotName, const std::string& logFile);

        virtual armarx::DistanceInfo getRobotNodeDistance(const std::string& robotName, const std::string& robotNodeName1, const std::string& robotNodeName2);
        virtual armarx::DistanceInfo getDistance(const std::string& robotName, const std::string& robotNodeName, const std::string& worldObjectName);

        std::vector<SimDynamics::DynamicsEngine::DynamicsContactInfo> copyContacts();


        virtual int getContactCont();

        virtual VirtualRobot::RobotPtr getRobot(const std::string& robotName);
        virtual std::map<std::string, VirtualRobot::RobotPtr> getRobots();
        virtual SimDynamics::DynamicsRobotPtr getDynamicRobot(const std::string& robotName);

        virtual std::vector<std::string> getRobotNames();
        virtual std::vector<std::string> getObstacleNames();

        virtual void setRobotNodeSimType(const std::string& robotName, const std::string& robotNodeName, VirtualRobot::SceneObject::Physics::SimulationType simType);
        virtual void setObjectSimType(const std::string& objectName, VirtualRobot::SceneObject::Physics::SimulationType simType);

    protected:
        virtual bool addObstacleEngine(VirtualRobot::SceneObjectPtr o, VirtualRobot::SceneObject::Physics::SimulationType simType);
        virtual bool removeObstacleEngine(const std::string& name);
        virtual bool addRobotEngine(VirtualRobot::RobotPtr robot, double pid_p, double pid_i, double pid_d, const std::string& filename, bool staticRobot, bool selfCollisions);

        //! create a joint
        virtual bool objectGraspedEngine(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName, Eigen::Matrix4f& storeLocalTransform);

        //! remove a joint
        virtual bool objectReleasedEngine(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName);

        virtual bool removeRobotEngine(const std::string& robotName);
        virtual bool synchronizeSimulationDataEngine();

        virtual void setupFloorEngine(bool enable, const std::string& floorTexture, float friction = -1);

        virtual bool synchronizeObjects();

        virtual bool getRobotStatus(const std::string& robotName,
                                    NameValueMap& jointAngles,
                                    NameValueMap& jointVelocities,
                                    NameValueMap& jointTorques,
                                    Eigen::Vector3f& linearVelocity,
                                    Eigen::Vector3f& angularVelocity);

        virtual bool updateForceTorqueSensor(ForceTorqueInfo& ftInfo);

        virtual bool synchronizeSceneObjectPoses(VirtualRobot::SceneObjectPtr currentObjEngine, std::map< std::string, armarx::PoseBasePtr >& objMap);

        virtual VirtualRobot::SceneObjectPtr getFloor();

        // manually apply velocities to static robots
        void stepStaticRobots(double deltaInSeconds);

        // we assume to have a bullet engine
        SimDynamics::BulletEnginePtr bulletEngine;
        int bulletFixedTimeStepMaxNrLoops;
        float bulletFixedTimeStepMS;
        int stepSizeMs;

        std::vector<SimDynamics::DynamicsEngine::DynamicsContactInfo> contacts;

        btClock m_clock;

        struct DynamicsRobotInfo
        {
            VirtualRobot::RobotPtr robot;
            SimDynamics::DynamicsRobotPtr dynamicsRobot;
            bool isStatic;
            std::map< std::string, float> targetVelocities; // only needed for static roobts
        };

        SimDynamics::DynamicsWorldPtr dynamicsWorld;
        std::map<std::string, DynamicsRobotInfo> dynamicRobots;
        std::vector<SimDynamics::DynamicsObjectPtr> dynamicsObjects;

        SimDynamics::BulletRobotLoggerPtr robotLogger;

        btScalar getDeltaTimeMicroseconds();

        SimDynamics::DynamicsObjectPtr getFirstDynamicsObject(const std::string& robotName, const std::string& nodeName);

        virtual bool synchronizeRobotNodePoses(const std::string& robotName, std::map<std::string, armarx::PoseBasePtr>& objMap);

        // SimulatedWorld interface
    public:
        virtual void setFixedTimeStepMS(int timeStepMS);
        virtual float getInternalTimeStepMS();
        virtual void setInternalTimeStepMS(float timeStepMS);
        virtual void deactivateCollisions() override;
    };

    typedef boost::shared_ptr<BulletPhysicsWorld> BulletPhysicsWorldPtr;
}

#endif

