/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::Simulator
 * @author     Nikolaus Vahrenkamp ( vahrenkamp at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef __ARMARX_KINEMATICS_WORLD__
#define __ARMARX_KINEMATICS_WORLD__

#include <VirtualRobot/VirtualRobotCommon.h>

#include <boost/thread/recursive_mutex.hpp>

#include <Eigen/Core>
#include <Eigen/Geometry>

// just needed for SceneData
#include <ArmarXSimulation/interface/simulator/SimulatorInterface.h>
#include <ArmarXCore/core/system/Synchronization.h>
#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <ArmarXCore/observers/filters/DerivationFilter.h>
#include <ArmarXCore/observers/filters/DerivationFilter.h>

#include "SimulatedWorld.h"

namespace armarx
{
    /*!
     * \brief The KinematicsWorld class encapsulates the kinemtics simulation and the corresponding data.
     * All real physics simulation is either approximated through simple kinemtics or disabled.
     *
     */
    class KinematicsWorld : public SimulatedWorld
    {
    public:

        KinematicsWorld();
        ~KinematicsWorld();

        virtual void initialize(int stepSizeMS, bool maxRealTimeSimSpeed = true, bool floorPlane = false, std::string floorTexture = std::string());

        //! ignoring velocity target
        virtual void actuateRobotJoints(const std::string& robotName, const std::map< std::string, float>& angles, const std::map< std::string, float>& velocities);
        virtual void actuateRobotJointsPos(const std::string& robotName, const std::map< std::string, float>& angles);
        virtual void actuateRobotJointsVel(const std::string& robotName, const std::map< std::string, float>& velocities);
        virtual void actuateRobotJointsTorque(const std::string& robotName, const std::map< std::string, float>& torques);
        virtual void setRobotPose(const std::string& robotName, const Eigen::Matrix4f& globalPose);

        virtual void applyForceRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& force);
        virtual void applyImpulseRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& impulse);
        virtual void applyTorqueRobotNode(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& torque);

        virtual void applyForceObject(const std::string& objectName, const Eigen::Vector3f& force);
        virtual void applyTorqueObject(const std::string& objectName, const Eigen::Vector3f& torque);

        virtual bool hasObject(const std::string& instanceName);
        virtual void setObjectPose(const std::string& objectName, const Eigen::Matrix4f& globalPose);

        virtual bool hasRobot(const std::string& robotName);
        virtual bool hasRobotNode(const std::string& robotName, const std::string& robotNodeName);
        virtual float getRobotMass(const std::string& robotName);

        virtual std::map< std::string, float> getRobotJointAngles(const std::string& robotName);
        virtual float getRobotJointAngle(const std::string& robotName, const std::string& nodeName);
        virtual std::map< std::string, float> getRobotJointVelocities(const std::string& robotName);
        virtual float getRobotJointVelocity(const std::string& robotName, const std::string& nodeName);
        virtual std::map< std::string, float> getRobotJointTorques(const std::string&);
        virtual ForceTorqueDataSeq getRobotForceTorqueSensors(const std::string&);

        virtual float getRobotJointLimitLo(const std::string& robotName, const std::string& nodeName);
        virtual float getRobotJointLimitHi(const std::string& robotName, const std::string& nodeName);
        virtual Eigen::Matrix4f getRobotPose(const std::string& robotName);

        virtual float getRobotMaxTorque(const std::string& robotName, const std::string& nodeName);
        virtual void setRobotMaxTorque(const std::string& robotName, const std::string& nodeName, float maxTorque);

        virtual Eigen::Matrix4f getRobotNodePose(const std::string& robotName, const std::string& robotNodeName);

        virtual Eigen::Vector3f getRobotLinearVelocity(const std::string& robotName, const std::string& nodeName);
        virtual Eigen::Vector3f getRobotAngularVelocity(const std::string& robotName, const std::string& nodeName);

        // the robotNodeName is ignored, the whole robot is moved...
        virtual void setRobotLinearVelocity(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotAngularVelocity(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotLinearVelocityRobotRootFrame(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);
        virtual void setRobotAngularVelocityRobotRootFrame(const std::string& robotName, const std::string& robotNodeName, const Eigen::Vector3f& vel);

        virtual Eigen::Matrix4f getObjectPose(const std::string& objectName);

        virtual VirtualRobot::RobotPtr getRobot(const std::string& robotName);
        virtual std::map<std::string, VirtualRobot::RobotPtr> getRobots();
        virtual VirtualRobot::SceneObjectPtr getObject(const std::string& objectName);
        virtual std::vector<VirtualRobot::SceneObjectPtr> getObjects();

        std::vector<SimDynamics::DynamicsEngine::DynamicsContactInfo> copyContacts();


        virtual FramedPosePtr toFramedPose(const Eigen::Matrix4f& globalPose, const std::string& robotName, const std::string& frameName);

        /**
        * Perform one simulation step. Calculates the delta update time from the time that has passed since the last call.
        * This updates all models.
        */
        virtual void stepPhysicsRealTime();

        /**
        * Perform one simulation step.
        * This updates all models.
        */
        virtual void stepPhysicsFixedTimeStep();

        /*!
         * \brief
         * \return The number of steps * the timestep in MS
         */
        virtual int getFixedTimeStepMS();


        virtual void enableLogging(const std::string& robotName, const std::string& logFile);

        virtual armarx::DistanceInfo getRobotNodeDistance(const std::string& robotName, const std::string& robotNodeName1, const std::string& robotNodeName2);
        virtual armarx::DistanceInfo getDistance(const std::string& robotName, const std::string& robotNodeName, const std::string& worldObjectName);

        virtual std::vector<std::string> getRobotNames();
        virtual std::vector<std::string> getObstacleNames();

        virtual void setRobotNodeSimType(const std::string& robotName, const std::string& robotNodeName, VirtualRobot::SceneObject::Physics::SimulationType simType);
        virtual void setObjectSimType(const std::string& objectName, VirtualRobot::SceneObject::Physics::SimulationType simType);

        virtual void deactivateCollisions() override;

    protected:
        virtual bool addObstacleEngine(VirtualRobot::SceneObjectPtr o, VirtualRobot::SceneObject::Physics::SimulationType simType);
        virtual bool removeObstacleEngine(const std::string& name);
        virtual bool addRobotEngine(VirtualRobot::RobotPtr robot, double pid_p, double pid_i, double pid_d, const std::string& filename, bool staticRobot, bool selfCollisions);

        //! create a joint
        virtual bool objectGraspedEngine(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName, Eigen::Matrix4f& storeLocalTransform);

        //! remove a joint
        virtual bool objectReleasedEngine(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName);

        virtual bool removeRobotEngine(const std::string& robotName);
        virtual bool synchronizeSimulationDataEngine();

        virtual void setupFloorEngine(bool enable, const std::string& floorTexture, float friction = -1);

        virtual bool synchronizeObjects();

        virtual bool getRobotStatus(const std::string& robotName,
                                    NameValueMap& jointAngles,
                                    NameValueMap& jointVelocities,
                                    NameValueMap& jointTorques,
                                    Eigen::Vector3f& linearVelocity,
                                    Eigen::Vector3f& angularVelocity);

        virtual bool updateForceTorqueSensor(ForceTorqueInfo& ftInfo);

        virtual bool synchronizeSceneObjectPoses(VirtualRobot::SceneObjectPtr currentObjEngine, std::map< std::string, armarx::PoseBasePtr >& objMap);

        virtual VirtualRobot::SceneObjectPtr getFloor();

        // manually apply velocities to static robots
        void stepStaticRobots(double deltaInSeconds);
        void stepStaticObjects(double deltaInSeconds);

        int stepSizeMs;

        struct KinematicRobotInfo
        {
            KinematicRobotInfo()
            {
                velTransTarget.setZero();
                velRotTarget.setZero();
                velTransActual.setZero();
                velRotActual.setZero();
            }

            VirtualRobot::RobotPtr robot;
            std::map< std::string, float> targetVelocities;
            std::map< std::string, float> actualVelocities;
            Eigen::Vector3f velTransTarget, velTransActual;
            Eigen::Vector3f velRotTarget, velRotActual;
        };

        std::map<std::string, KinematicRobotInfo> kinematicRobots;
        std::vector<VirtualRobot::SceneObjectPtr> kinematicObjects;
        std::map<VirtualRobot::RobotNodePtr, std::pair<IceUtil::Time, float> > jointAngleDerivationFilters;
        std::map<VirtualRobot::RobotPtr, std::pair<IceUtil::Time, Eigen::Vector3f> > linearVelocityFilters;
        std::map<VirtualRobot::RobotPtr, std::pair<IceUtil::Time, Eigen::Matrix4f>> angularVelocityFilters;
        VirtualRobot::SceneObjectPtr floor;
        Eigen::Vector3f floorPos;
        Eigen::Vector3f floorUp;
        double floorExtendMM;
        double floorDepthMM;
        VirtualRobot::ObstaclePtr groundObject;
        virtual void createFloorPlane(const Eigen::Vector3f& pos, const Eigen::Vector3f& up);

        IceUtil::Time lastTime;

        float getDeltaTimeMilli();

        virtual bool synchronizeRobotNodePoses(const std::string& robotName, std::map<std::string, armarx::PoseBasePtr>& objMap);
        void calculateActualVelocities();

        // SimulatedWorld interface
    public:
        virtual void setFixedTimeStepMS(int timeStepMS);
        virtual float getInternalTimeStepMS();
        virtual void setInternalTimeStepMS(float timeStepMS);

    };

    typedef boost::shared_ptr<KinematicsWorld> KinematicsWorldPtr;
}

#endif

