/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2013-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::Simulator
 * @author     Nikolaus ( vahrenkamp at kit dot edu )
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// print who is querying for locks (scoped lock queries are counted)
//#define PERFORMANCE_EVALUATION_LOCKS

#define MAX_SIM_TIME_WARNING 25
#define FORCE_TORQUE_TOPIC_NAME "ForceTorqueDynamicSimulationValues"

#include "SimulatedWorld.h"

#include <VirtualRobot/Nodes/ForceTorqueSensor.h>
#include <VirtualRobot/CollisionDetection/CollisionChecker.h>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <algorithm>

using namespace VirtualRobot;
using namespace armarx;

//#define PRINT_TEST_DEBUG_MESSAGES

SimulatedWorld::SimulatedWorld()
{
    engineMutex.reset(new boost::recursive_mutex());
    synchronizeMutex.reset(new boost::recursive_mutex());
    simStepExecutionTimeMS = 0.0f;
    synchronizeTimeMS = 0.0f;
    simTimeStepMS = 0.0f;
    currentSimTimeSec = 0.0f;
    collectContacts = false;
}

SimulatedWorld::~SimulatedWorld()
{
}

void SimulatedWorld::activateObject(const std::string& objectName)
{

}

bool SimulatedWorld::resetData()
{
    ARMARX_INFO << "Removing all robots & objects";
    bool res = true;
    res &= removeRobots();
    res &= removeObstacles();

    return res;
}

double SimulatedWorld::getCurrentSimTime()
{
    return currentSimTimeSec;
}

void SimulatedWorld::resetSimTime()
{
    currentSimTimeSec = 0;
}

bool SimulatedWorld::removeRobots()
{
    ARMARX_INFO << "Removing all robots";

    bool res = true;

    std::vector< std::string > names = getRobotNames();
    for (auto n : names)
    {
        res &= removeRobot(n);
    }

    return res;
}


SimulatedWorldData& SimulatedWorld::getReportData()
{
    return simReportData;
}

SimulatedWorldData SimulatedWorld::copyReportData()
{
    ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);
    return simReportData;
}


bool SimulatedWorld::addRobot(RobotPtr robot, double pid_p, double pid_i, double pid_d, const std::string& filename, bool staticRobot, float scaling, bool colModel, bool selfCollisions)
{
    ARMARX_VERBOSE_S << "Adding robot " << robot->getName() << ", file:" << filename;
    assert(robot);

    // lock engine and data
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);

    if (scaling != 1.0f)
    {
        ARMARX_INFO_S << "Scaling robot, factor: " << scaling;
        robot = robot->clone(robot->getName(), robot->getCollisionChecker(), scaling);
    }


    bool ok = addRobotEngine(robot, pid_p, pid_i, pid_d, filename, staticRobot, selfCollisions);
    if (!ok)
    {
        return false;
    }

    RobotVisuData robData;

    robData.name = robot->getName();
    robData.robotFile = filename;
    robData.updated = true;
    robData.scaling = scaling;
    robData.colModel = colModel;

    simVisuData.robots.push_back(robData);

    RobotInfo robInfo;
    robInfo.robotName = robot->getName();

    try
    {
        // setup robot report data
        robInfo.robotTopicName = "Simulator_Robot_";
        robInfo.robotTopicName += robot->getName();

        std::vector<SensorPtr> sensors = robot->getSensors();

        for (size_t i = 0; i < sensors.size(); i++)
        {
            ForceTorqueSensorPtr ft = boost::dynamic_pointer_cast<ForceTorqueSensor>(sensors[i]);

            if (ft)
            {
                ForceTorqueInfo fti;
                fti.enable = true;
                fti.robotName = robot->getName();
                fti.sensorName = ft->getName();
                fti.ftSensor = ft;
                fti.robotNodeName = ft->getRobotNode()->getName();
                fti.topicName = FORCE_TORQUE_TOPIC_NAME;

                robInfo.forceTorqueSensors.push_back(fti);
            }
        }
    }
    catch (...)
    {
        ARMARX_WARNING_S << "addRobot failed";
    }

    simReportData.robots.push_back(robInfo);

    synchronizeRobots();
    return true;
}



void SimulatedWorld::objectReleased(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName)
{
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ARMARX_INFO_S << "Object released, robot " << robotName << ", node:" << robotNodeName << ", object:" << objectName;

    if (objectReleasedEngine(robotName, robotNodeName, objectName))
    {
        // check if object is in  grasped list
        auto aoIt = attachedObjects.begin();
        bool objectFound = false;

        while (aoIt != attachedObjects.end())
        {
            if (aoIt->robotName == robotName && aoIt->robotNodeName == robotNodeName && aoIt->objectName == objectName)
            {
                attachedObjects.erase(aoIt);
                objectFound = true;
                break;
            }

            aoIt++;
        }

        if (!objectFound)
        {
            ARMARX_WARNING_S << "Object was not attached to robot: " << objectName;
        }

        ARMARX_INFO_S << "Successfully detached object " << objectName;
    }
}

void SimulatedWorld::objectGrasped(const std::string& robotName, const std::string& robotNodeName, const std::string& objectName)
{
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ARMARX_INFO_S << "Object grasped, robot " << robotName << ", node:" << robotNodeName << ", object:" << objectName;

    // check if object is already grasped
    for (const auto& ao : attachedObjects)
    {
        if (ao.robotName == robotName && ao.robotNodeName == robotNodeName && ao.objectName == objectName)
        {
            ARMARX_INFO_S << "Object already attached, skipping...";
            return;
        }
    }

    Eigen::Matrix4f localTransform;
    if (objectGraspedEngine(robotName, robotNodeName, objectName, localTransform))
    {
        ARMARX_INFO_S << "Successfully attached object " << objectName;
        // store grasping information
        GraspingInfo g;
        g.robotName = robotName;
        g.robotNodeName = robotNodeName;
        g.objectName = objectName;

        g.node2objectTransformation = localTransform;//rn->toLocalCoordinateSystem(o->getSceneObject()->getGlobalPose());

        attachedObjects.push_back(g);
    }
}


bool SimulatedWorld::removeObstacles()
{
    ARMARX_VERBOSE_S << "Removing all obstacles";
    bool result = true;
    std::vector<std::string> names = getObstacleNames();

    for (auto& n : names)
    {
        if (!removeObstacle(n))
        {
            ARMARX_WARNING_S << "Could not remove object with name " << n;
            result = false;
        }
    }

    return result;
}


bool SimulatedWorld::removeObstacle(const std::string& name)
{
    ARMARX_VERBOSE_S << "Removing obstacle " << name;
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockData = getScopedSyncLock(__FUNCTION__);

    if (!removeObstacleEngine(name))
    {
        return false;
    }

    {
        // access data
        bool objectFound = false;

        for (auto s = simVisuData.objects.begin(); s != simVisuData.objects.end(); s++)
        {
            if (s->name == name)
            {
                objectFound = true;
                simVisuData.objects.erase(s);
                break;
            }
        }

        if (!objectFound)
        {
            ARMARX_WARNING_S << "Object with name " << name << " not in synchronized list, cannot remove";
        }
    }

    return true;
}


bool SimulatedWorld::addRobot(const std::string& robotInstanceName, const std::string& filename, Eigen::Matrix4f pose, const std::string& filenameLocal, double pid_p, double pid_i, double pid_d, bool staticRobot, float scaling, bool colModel, std::map<std::string, float> initConfig, bool selfCollisions)
{
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ARMARX_INFO_S << "Loading Robot from " << filename << endl;
    RobotPtr robot;

    try
    {
        VirtualRobot::RobotIO::RobotDescription loadMode = VirtualRobot::RobotIO::eFull;
        if (colModel)
        {
            loadMode = VirtualRobot::RobotIO::eCollisionModel;
        }
        robot = RobotIO::loadRobot(filename, loadMode);

        // ensure that the robot has the standard name

        robot->setName(robotInstanceName.empty() ? robot->getType() : robotInstanceName);
    }
    catch (VirtualRobotException& e)
    {
        ARMARX_ERROR_S << " ERROR while creating robot (file:" << filename << ")" << endl << e.what() << endl;
        return false;
    }

    if (!robot)
    {
        ARMARX_ERROR_S << " ERROR while creating robot (file:" << filename << ")" << endl;
        return false;
    }

    ARMARX_INFO_S << "Loaded robot with name " << robot->getName();

    /*if (initConfig.size() > 0)
    {
        robot->setJointValues(initConfig);
    }*/

    if (pose.isIdentity() && getFloor())
    {
        VirtualRobot::BoundingBox bbox = robot->getBoundingBox();
        //pose(2,3) = -bbox.getMin()(2) + 1.0f;

        // move down robot until collision
        std::vector<CollisionModelPtr> colModels = robot->getCollisionModels();
        CollisionModelPtr floorCol = getFloor()->getCollisionModel();

        bool found = false;
        bool loop = 0;
        float z = -bbox.getMin()(2) + 20.0f;
        pose(2, 3) = z;
        robot->setGlobalPose(pose);
        found = (robot->getCollisionChecker()->checkCollision(colModels, floorCol));

        while (loop < 1000 && !found)
        {
            z -= 2.0f;
            pose(2, 3) = z;
            robot->setGlobalPose(pose);
            found = (robot->getCollisionChecker()->checkCollision(colModels, floorCol));
            loop++;
        }

        if (found)
        {
            // move up
            z += 2.0f;
            pose(2, 3) = z;
            robot->setGlobalPose(pose);
        }
    }

    robot->setGlobalPose(pose);
    std::string fn = filename;

    if (!filenameLocal.empty())
    {
        fn = filenameLocal;
    }

    bool success = addRobot(robot, pid_p, pid_i, pid_d, fn, staticRobot, scaling, colModel, selfCollisions);

    if (success && initConfig.size() > 0)
    {
        actuateRobotJointsPos(robot->getName(), initConfig);
    }

    return success;
}


bool SimulatedWorld::removeRobot(const std::string& robotName)
{
    ARMARX_INFO << "Removing robot " << robotName;

    // lock engine and data
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);

    bool ok = removeRobotEngine(robotName);
    if (!ok)
    {
        return false;
    }

    // access data
    {
        //        if (simVisuData.robots.size() != 1 || simVisuData.robots.at(0).name != robotName)
        //        {
        //            ARMARX_WARNING_S << "Internal data structure wrong...";
        //        }

        simVisuData.robots.clear();
        simReportData.robots.clear();
    }

    // check if there are grasped objects
    auto aoIt = attachedObjects.begin();

    while (aoIt != attachedObjects.end())
    {
        if (aoIt->robotName == robotName)
        {
            ARMARX_DEBUG << "Removing attached object " << aoIt->objectName;
            aoIt = attachedObjects.erase(aoIt);
        }
        else
        {
            aoIt++;
        }
    }

    return true;
}

bool SimulatedWorld::addScene(const std::string& filename, VirtualRobot::SceneObject::Physics::SimulationType simType)
{
    ARMARX_INFO_S << "Loading Scene from " << filename << endl;
    ScenePtr scene;

    try
    {
        scene = SceneIO::loadScene(filename);
    }
    catch (VirtualRobotException& e)
    {
        ARMARX_ERROR_S << " ERROR while creating scene (file:" << filename << ")" << endl << e.what() << endl;
        return false;
    }

    if (!scene)
    {
        cout << " ERROR while creating scene (file:" << filename << ")" << endl;
        return false;
    }

    return addScene(scene, simType);
}

bool SimulatedWorld::addScene(ScenePtr scene, VirtualRobot::SceneObject::Physics::SimulationType simType)
{
    if (!scene)
    {
        ARMARX_ERROR_S << "No Scene!" << endl;
        return false;
    }

    int count = 0;
    std::vector<ManipulationObjectPtr> mo = scene->getManipulationObjects();

    for (size_t i = 0; i < mo.size(); i++)
    {
        ARMARX_VERBOSE_S << "Scene: ManipulationObject " << mo[i]->getName();
        addObstacle(mo[i], simType, mo[i]->getFilename());
        count++;
    }

    std::vector<ObstaclePtr> o = scene->getObstacles();

    for (size_t i = 0; i < o.size(); i++)
    {
        ARMARX_VERBOSE_S << "Scene: Obstacle " << o[i]->getName();
        addObstacle(o[i], simType, o[i]->getFilename());
        count++;
    }

    int countrobot = 0;
    std::vector<RobotPtr> ro = scene->getRobots();

    for (size_t i = 0; i < ro.size(); i++)
    {
        // ensure consistent name
        ro[i]->setName(ro[i]->getType());
        ARMARX_VERBOSE_S << "Scene: Robot " << ro[i]->getName() << ", type:" << ro[i]->getType();
        addRobot(ro[i], 10.0f, 0.0, 0.0, ro[i]->getFilename());
        countrobot++;
    }

    ARMARX_INFO_S << " Added " << countrobot << " robots and " << count << " objects" << endl;

    return true;
}

bool SimulatedWorld::addObstacle(const std::string& filename, Eigen::Matrix4f pose, VirtualRobot::SceneObject::Physics::SimulationType simType, const std::string& localFilename)
{
    ARMARX_INFO_S << "Loading obstacle from " << filename << ", (local file:" << localFilename << ")";
    ObstaclePtr o;

    try
    {
        o = ObjectIO::loadObstacle(filename);
    }
    catch (VirtualRobotException& e)
    {
        ARMARX_ERROR_S << " ERROR while creating obstacle (file:" << filename << ")" << endl << e.what() << endl;
        return false;
    }

    if (!o)
    {
        cout << " ERROR while creating (file:" << filename << ")" << endl;
        return false;
    }

    o->setGlobalPose(pose);
    std::string fn = filename;

    if (!localFilename.empty())
    {
        std::string absoluteFilename;
        // check if we would find the local file
        bool fileFound = ArmarXDataPath::getAbsolutePath(localFilename, absoluteFilename);

        if (fileFound)
        {
            ARMARX_IMPORTANT_S << "found local path:" << localFilename << " -> " << absoluteFilename;
            fn = localFilename;
        }
        else
        {
            ARMARX_IMPORTANT_S << "could not find global path from local path:" << localFilename << " -> usign " << fn;
        }
    }

    return addObstacle(o, simType, fn);
}


bool SimulatedWorld::addObstacle(VirtualRobot::SceneObjectPtr o, VirtualRobot::SceneObject::Physics::SimulationType simType, const std::string& filename, const std::string& objectClassName, ObjectVisuPrimitivePtr primitiveData)
{
    if (!o)
    {
        return false;
    }

    ARMARX_VERBOSE_S << "Adding obstacle " << o->getName();

    // lock engine and data
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockData = getScopedSyncLock(__FUNCTION__);

    //VirtualRobot::SceneObjectPtr clonedObj;
    if (hasObject(o->getName()))
    {
        ARMARX_ERROR_S << "Object with name " << o->getName() << " already present! Choose a different instance name";
        return false;
    }

    bool ok = addObstacleEngine(o, simType);
    if (!ok)
    {
        return false;
    }

    {
        // access data

        if (primitiveData && (!filename.empty() || !objectClassName.empty()))
        {
            ARMARX_WARNING_S << "Internal error, either filename or classname or primitive";
        }

        armarx::ObjectVisuData ovd;
        ovd.name = o->getName();
        ovd.filename = filename;
        ovd.objectPoses[o->getName()] = new Pose(o->getGlobalPose());
        ovd.objectClassName = objectClassName;

        if (primitiveData)
        {
            ovd.objectPrimitiveData = primitiveData;
        }

        ovd.updated = true;
        simVisuData.objects.push_back(ovd);

    }

    return true;
}


ScopedRecursiveLockPtr SimulatedWorld::getScopedEngineLock(const std::string& callStr)
{
    ScopedRecursiveLockPtr l(new ScopedRecursiveLock(*engineMutex));
#ifdef PERFORMANCE_EVALUATION_LOCKS
    if (engineMtxAccTime.find(callStr) == engineMtxAccTime.end())
    {
        engineMtxAccCalls[callStr] = 1;
        engineMtxAccTime[callStr] = 0;
        engineMtxLastTime[callStr] = IceUtil::Time::now();
    }
    else
    {
        IceUtil::Time delta = IceUtil::Time::now() - engineMtxLastTime[callStr];
        engineMtxAccCalls[callStr] += 1;
        engineMtxAccTime[callStr] += delta.toMilliSecondsDouble();
        engineMtxLastTime[callStr] = IceUtil::Time::now();
        if (engineMtxAccTime[callStr] > 1000)
        {
            ARMARX_INFO << "engine mutex performance [" << callStr << "]: Nr Calls:" << engineMtxAccCalls[callStr] << ", Calls/Sec:" << (float)engineMtxAccCalls[callStr] / engineMtxAccTime[callStr] * 1000.0f;
            engineMtxAccCalls[callStr] = 1;
            engineMtxAccTime[callStr] = 0;
            engineMtxLastTime[callStr] = IceUtil::Time::now();
        }
    }
#endif
    return l;
}

ScopedRecursiveLockPtr SimulatedWorld::getScopedSyncLock(const std::string& callStr)
{
    ScopedRecursiveLockPtr l(new ScopedRecursiveLock(*synchronizeMutex));
#ifdef PERFORMANCE_EVALUATION_LOCKS
    if (syncMtxAccTime.find(callStr) == syncMtxAccTime.end())
    {
        syncMtxAccCalls[callStr] = 1;
        syncMtxAccTime[callStr] = 0;
        syncMtxLastTime[callStr] = IceUtil::Time::now();
    }
    else
    {
        IceUtil::Time delta = IceUtil::Time::now() - syncMtxLastTime[callStr];
        syncMtxAccCalls[callStr] += 1;
        syncMtxAccTime[callStr] += delta.toMilliSecondsDouble();
        syncMtxLastTime[callStr] = IceUtil::Time::now();
        if (syncMtxAccTime[callStr] > 1000)
        {
            ARMARX_INFO << "snyc mutex performance [" << callStr << "]: Nr Calls:" << syncMtxAccCalls[callStr] << ", Calls/Sec:" << (float)syncMtxAccCalls[callStr] / syncMtxAccTime[callStr] * 1000.0f;
            syncMtxAccCalls[callStr] = 1;
            syncMtxAccTime[callStr] = 0;
            syncMtxLastTime[callStr] = IceUtil::Time::now();
        }
    }
#endif
    return l;
}


void SimulatedWorld::setupFloor(bool enable, const std::string& floorTexture, float friction)
{
    setupFloorEngine(enable, floorTexture, friction);

    {
        ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);
        simVisuData.floor = enable;
        simVisuData.floorTextureFile = floorTexture;
    }
}

float SimulatedWorld::getSimulationStepDuration()
{
    return simTimeStepMS;
}

float SimulatedWorld::getSimulationStepTimeMeasured()
{
    return simStepExecutionTimeMS;
}

void SimulatedWorld::enableLogging(const std::string& robotName, const std::string& logFile)
{

}

float SimulatedWorld::getSimTime()
{
    return simStepExecutionTimeMS;
}

float SimulatedWorld::getSyncEngineTime()
{
    return synchronizeTimeMS;
}


bool SimulatedWorld::synchronizeRobots()
{
    // lock engine and data
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);

    // update visu data
    if (simVisuData.robots.size() == 0)
    {
        return false;
    }

    for (size_t robotNr = 0; robotNr < simVisuData.robots.size(); robotNr++)
    {
        std::string name = simVisuData.robots.at(robotNr).name;

        if (!synchronizeRobotNodePoses(name, simVisuData.robots.at(robotNr).robotNodePoses))
        {
            ARMARX_ERROR_S << "Failed to synchronize robot..." << endl;
            return false;
        }

        armarx::PosePtr p(new Pose(getRobotPose(name)));
        simVisuData.robots.at(robotNr).pose = p;
        simVisuData.robots.at(robotNr).updated = true;
    }

    // update data for reporting

    for (size_t robotNr = 0; robotNr < simReportData.robots.size(); robotNr++)
    {
        std::string name = simReportData.robots.at(robotNr).robotName;

        simReportData.robots.at(robotNr).pose = getRobotPose(name);

        getRobotStatus(name,
                       simReportData.robots.at(robotNr).jointAngles,
                       simReportData.robots.at(robotNr).jointVelocities,
                       simReportData.robots.at(robotNr).jointTorques,
                       simReportData.robots.at(robotNr).linearVelocity,
                       simReportData.robots.at(robotNr).angularVelocity
                      );


        // update force torque info
        for (size_t i = 0; i < simReportData.robots.at(robotNr).forceTorqueSensors.size(); i++)
        {
            updateForceTorqueSensor(simReportData.robots.at(robotNr).forceTorqueSensors[i]);
        }
    }

    return true;
}


bool SimulatedWorld::synchronizeSimulationData()
{
    IceUtil::Time startTime = IceUtil::Time::now();

    // lock engine and data
    ScopedRecursiveLockPtr lockEngine = getScopedEngineLock(__FUNCTION__);
    ScopedRecursiveLockPtr lockSync = getScopedSyncLock(__FUNCTION__);

    IceUtil::Time durationlock = IceUtil::Time::now() - startTime;

    if (durationlock.toMilliSecondsDouble() > 4)
    {
        ARMARX_INFO << deactivateSpam(10) << "Engine/Sync locking took long :" << durationlock.toMilliSecondsDouble() << " ms";
    }

    simVisuData.timestamp = getCurrentSimTime() * 1000;
    simReportData.timestamp = IceUtil::Time::milliSeconds(simVisuData.timestamp);

    IceUtil::Time startTimeR = IceUtil::Time::now();
    // ROBOT -> update visu & report data
    synchronizeRobots();
    IceUtil::Time durationR = IceUtil::Time::now() - startTimeR;
    if (durationR.toMilliSecondsDouble() > 4)
    {
        ARMARX_INFO << deactivateSpam(10) << "Robot sync took long:" << durationR.toMilliSecondsDouble() << " ms";
    }

    IceUtil::Time startTimeO = IceUtil::Time::now();
    // OBJECTS -> visu & report data update
    synchronizeObjects();
    IceUtil::Time durationO = IceUtil::Time::now() - startTimeO;
    if (durationO.toMilliSecondsDouble() > 4)
    {
        ARMARX_INFO << deactivateSpam(10) << "Object sync took long:" << durationO.toMilliSecondsDouble() << " ms";
    }

    // CONTACTS
    synchronizeSimulationDataEngine();

    IceUtil::Time duration = IceUtil::Time::now() - startTime;
    synchronizeTimeMS = (float)duration.toMilliSecondsDouble();

    return true;
}

void SimulatedWorld::updateContacts(bool enable)
{
    collectContacts = enable;
}


SceneVisuData SimulatedWorld::copySceneVisuData()
{
    ScopedRecursiveLockPtr lock = getScopedSyncLock(__FUNCTION__);
    return simVisuData;
}


int SimulatedWorld::getRobotJointAngleCount()
{
    ScopedRecursiveLockPtr l = getScopedSyncLock(__FUNCTION__);

    if (simVisuData.robots.size() == 0)
    {
        return 0;
    }

    return simVisuData.robots.at(0).robotNodePoses.size();
}

int SimulatedWorld::getContactCont()
{
    ScopedRecursiveLockPtr l = getScopedSyncLock(__FUNCTION__);
    return 0;
}

int SimulatedWorld::getObjectCount()
{
    ScopedRecursiveLockPtr l = getScopedSyncLock(__FUNCTION__);
    return simVisuData.objects.size();
}



